from io import StringIO
from json import dumps as json_dumps
import pathlib
from subprocess import list2cmdline
from typing import List, Optional, Tuple

import click
from rich.console import Console
import yaml

import anyscale
from anyscale._private.models.image_uri import ImageURI
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.commands import command_examples
from anyscale.commands.util import (
    AnyscaleCommand,
    build_kv_table,
    convert_kv_strings_to_dict,
    override_env_vars,
    parse_repeatable_tags_to_dict,
    parse_tags_kv_to_str_map,
)
from anyscale.controllers.job_controller import JobController
from anyscale.job.models import JobConfig, JobLogMode, JobState, JobStatus
from anyscale.util import validate_list_jobs_state_filter, validate_non_negative_arg


log = BlockLogger()  # CLI Logger


def _validate_job_name_and_id(name: Optional[str], id: Optional[str]):  # noqa: A002
    if name is None and id is None:
        raise click.ClickException("One of '--name' and '--id' must be provided.")

    if name is not None and id is not None:
        raise click.ClickException("Only one of '--name' and '--id' can be provided.")


def _check_for_new_format_fields(config_file: str) -> None:
    """Check if a config file contains new-format fields that require using -f flag.

    Raises a ClickException if new-format fields are detected, suggesting the user
    should use the --config-file/-f flag instead.
    """
    # Fields that are specific to the new job submission API
    NEW_FORMAT_FIELDS = {
        "image_uri",
        "containerfile",
        "working_dir",
        "requirements",
        "env_vars",
        "py_modules",
        "excludes",
        "ray_version",
        "registry_login_secret",
        "timeout_s",
        "tags",
    }

    try:
        with open(config_file) as f:
            config_dict = yaml.safe_load(f) or {}
    except Exception:  # noqa: BLE001
        # If we can't read the file, let the normal flow handle the error
        return

    if not isinstance(config_dict, dict):
        return

    found_fields = [field for field in NEW_FORMAT_FIELDS if field in config_dict]

    if found_fields:
        fields_str = ", ".join(f"'{field}'" for field in found_fields)
        raise click.ClickException(
            f"Your config file contains fields that require the new job submission API: {fields_str}\n\n"
            f"Please use the '--config-file' or '-f' flag instead:\n"
            f"  anyscale job submit -f {config_file}\n\n"
            f"Alternatively, update your config to use the legacy format.\n"
            f"See https://docs.anyscale.com/reference/job-api/ for more information."
        )


@click.group("job", help="Interact with production jobs running on Anyscale.")
def job_cli() -> None:
    pass


@job_cli.command(
    name="submit",
    short_help="Submit a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_SUBMIT_EXAMPLE,
)
@click.option("-n", "--name", required=False, default=None, help="Name of the job.")
@click.option(
    "-w",
    "--wait",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Block this CLI command and print logs until the job finishes.",
)
@click.option(
    "--config-file",
    "-f",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to use for this job. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the job. This defaults to the compute configuration of the workspace.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for this job. When running in a workspace, this defaults to the image of the workspace.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the job.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the job. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
@click.option(
    "--working-dir",
    required=False,
    default=None,
    type=str,
    help="Path to a local directory or a remote URI to a .zip file (S3, GS, HTTP) that will be the working directory for the job. The files in the directory will be automatically uploaded to cloud storage. When running in a workspace, this defaults to the current working directory.",
)
@click.option(
    "-e",
    "--exclude",
    required=False,
    type=str,
    multiple=True,
    help="File pattern to exclude when uploading local directories. This argument can be specified multiple times and the patterns will be appended to the 'excludes' list in the config file (if any).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the job. Anyscale installs these dependencies on top of the image. If you run a job from a workspace, the default is to use the workspace dependencies, but specifying this option overrides them.",
)
@click.option(
    "--py-module",
    required=False,
    default=None,
    multiple=True,
    type=str,
    help="Python modules to be available for import in the Ray workers. Each entry must be a path to a local directory.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If you don't specify a Ray version, Anyscale defaults to the latest Ray version available at the time of the Anyscale CLI/SDK release.",
)
@click.option(
    "--max-retries",
    required=False,
    default=None,
    type=int,
    help="Maximum number of retries to attempt before failing the entire job.",
)
@click.option(
    "--timeout-s",
    "--timeout",
    "-t",
    required=False,
    default=None,
    type=int,
    help="The timeout in seconds for each job run. Set to None for no limit to be set.",
)
@click.argument("entrypoint", required=False, nargs=-1, type=click.UNPROCESSED)
def submit(  # noqa: PLR0912 PLR0913 C901
    entrypoint: Tuple[str],
    name: Optional[str],
    wait: Optional[bool],
    config_file: Optional[str],
    compute_config: Optional[str],
    image_uri: Optional[str],
    ray_version: Optional[str],
    registry_login_secret: Optional[str],
    containerfile: Optional[str],
    env: Tuple[str],
    working_dir: Optional[str],
    exclude: Tuple[str],
    requirements: Optional[str],
    py_module: Tuple[str],
    tags: Tuple[str],
    cloud: Optional[str],
    project: Optional[str],
    max_retries: Optional[int],
    timeout_s: Optional[int],
):
    """Submit a job.

    The job config can be specified in one of the following ways:

    * Job config file can be specified as a single positional argument. E.g. `anyscale job submit config.yaml`.

    * Job config can also be specified with command-line arguments. In this case, the entrypoint should be specified
as the positional arguments starting with `--`. Other arguments can be specified with command-line flags. E.g.

      * `anyscale job submit -- python main.py`: submit a job with the entrypoint `python main.py`.

      * `anyscale job submit --name my-job -- python main.py`: submit a job with the name `my-job` and the
entrypoint `python main.py`.

    * [Experimental] If you want to specify a config file and override some arguments with the commmand-line flags,
use the `--config-file` flag. E.g.

      * `anyscale job submit --config-file config.yaml`: submit a job with the config in `config.yaml`.

      * `anyscale job submit --config-file config.yaml -- python main.py`: submit a job with the config in `config.yaml`
and override the entrypoint with `python main.py`.

    Either containerfile or image-uri should be used, specifying both will result in an error.

    By default, this command submits the job asynchronously and exits. To wait for the job to complete, use the `--wait` flag.
    """

    job_controller = JobController()
    if len(entrypoint) == 1 and (
        pathlib.Path(entrypoint[0]).is_file() or entrypoint[0].endswith(".yaml")
    ):
        # If entrypoint is a single string that ends with .yaml, e.g. `anyscale job submit config.yaml`,
        # treat it as a config file, and use the old job submission API.
        if config_file is not None:
            raise click.ClickException(
                "`--config-file` should not be used when providing a config file as the entrypoint."
            )
        if image_uri:
            raise click.ClickException(
                "`--image-uri` should not be used when providing a config file as the entrypoint."
            )
        if registry_login_secret:
            raise click.ClickException(
                "`--registry-login-secret` should not be used when providing a config file as the entrypoint."
            )
        if containerfile:
            raise click.ClickException(
                "`--containerfile` should not be used when providing a config file as the entrypoint."
            )
        if env:
            raise click.ClickException(
                "`--env` should not be used when providing a config file as the entrypoint."
            )

        config_file = entrypoint[0]
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Job config file '{config_file}' not found.")

        # Check if the config file contains new-format fields that require using -f
        _check_for_new_format_fields(config_file)

        log.info(f"Submitting job from config file {config_file}.")

        job_id = job_controller.submit(config_file, name=name)
    else:
        # Otherwise, use the new job submission API. E.g.
        # `anyscale job submit -- python main.py`,
        # or `anyscale job submit --config-file config.yaml`.
        if len(entrypoint) == 0 and config_file is None:
            raise click.ClickException(
                "Either a config file or an inlined entrypoint must be provided."
            )
        if config_file is not None and not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Job config file '{config_file}' not found.")

        args = {}
        if len(entrypoint) > 0:
            args["entrypoint"] = list2cmdline(entrypoint)
        if name:
            args["name"] = name

        if containerfile and image_uri:
            raise click.ClickException(
                "Only one of '--containerfile' and '--image-uri' can be provided."
            )

        if registry_login_secret and (
            not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
        ):
            raise click.ClickException(
                "Registry login secret can only be used with an image that is not hosted on Anyscale.",
            )

        if ray_version and (not image_uri and not containerfile):
            raise click.ClickException(
                "Ray version can only be used with an image or containerfile.",
            )

        if image_uri:
            args["image_uri"] = image_uri

        if registry_login_secret:
            args["registry_login_secret"] = registry_login_secret

        if containerfile:
            args["containerfile"] = containerfile

        if ray_version:
            args["ray_version"] = ray_version

        if working_dir:
            args["working_dir"] = working_dir

        if config_file is not None:
            config = JobConfig.from_yaml(config_file, **args)
        else:
            config = JobConfig.from_dict(args)

        if compute_config is not None:
            config = config.options(compute_config=compute_config)

        if exclude:
            config = config.options(excludes=[e for e in exclude])

        if requirements is not None:
            if not pathlib.Path(requirements).is_file():
                raise click.ClickException(
                    f"Requirements file '{requirements}' not found."
                )
            config = config.options(requirements=requirements)

        if env:
            config = override_env_vars(config, convert_kv_strings_to_dict(env))

        if py_module:
            for module in py_module:
                if not pathlib.Path(module).is_dir():
                    raise click.ClickException(
                        f"Python module path '{module}' does not exist or is not a directory."
                    )
            config = config.options(py_modules=[*py_module])

        if cloud is not None:
            config = config.options(cloud=cloud)
        if project is not None:
            config = config.options(project=project)

        if max_retries is not None:
            config = config.options(max_retries=max_retries)

        if timeout_s is not None:
            config = config.options(timeout_s=timeout_s)

        if tags:
            tag_map = parse_tags_kv_to_str_map(tags)
            if tag_map:
                config = config.options(tags=tag_map)

        log.info(f"Submitting job with config {config}.")
        job_id = anyscale.job.submit(config)

    if wait:
        log.info(
            "Waiting for the job to run. Interrupting this command will not cancel the job."
        )
        anyscale.job.wait(id=job_id, follow=True)
    else:
        log.info("Use `--wait` to wait for the job to run and stream logs.")


# TODO(mowen): Add cloud support for this when we refactor to new SDK method
@job_cli.command(
    name="list",
    help="Display information about existing jobs.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_LIST_EXAMPLE,
)
@click.option("--name", "-n", required=False, default=None, help="Filter by job name.")
@click.option(
    "--id", "--job-id", required=False, default=None, help="Filter by job id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--include-all-users",
    is_flag=True,
    default=False,
    help="Include jobs not created by current user.",
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help=(
        "List archived jobs as well as unarchived jobs."
        "If not provided, defaults to listing only unarchived jobs."
    ),
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help=(
        "This option can be repeated to filter by multiple tags. "
        "Tags with the same key are ORed, whereas tags with different keys are ANDed. "
        "Example: --tag team:mlops --tag team:infra --tag env:prod. "
        "Filters with team: (mlops OR infra) AND env:prod."
    ),
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
@click.option(
    "--state",
    "-s",
    "states",
    required=False,
    multiple=True,
    help=f"Filter jobs by state. Accepts one or more states. Allowed states: {', '.join(HaJobStates.allowable_values)}",
    callback=validate_list_jobs_state_filter,
)
def list(  # noqa: A001 PLR0913
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    project_id: Optional[str],
    include_all_users: bool,
    include_archived: bool,
    max_items: int,
    states: List[HaJobStates],
    tags: List[str],
) -> None:
    job_controller = JobController()
    job_controller.list(
        name=name,
        job_id=id,
        project_id=project_id,
        include_all_users=include_all_users,
        include_archived=include_archived,
        max_items=max_items,
        states=states,
        tags=parse_repeatable_tags_to_dict(tags) if tags else None,
    )


@job_cli.command(
    name="archive",
    short_help="Archive a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_ARCHIVE_EXAMPLE,
)
@click.option("--id", "--job-id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def archive(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Archive a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be archived.
    """
    _validate_job_name_and_id(name=name, id=id)
    anyscale.job.archive(id=id, name=name, cloud=cloud, project=project)


@job_cli.command(
    name="terminate",
    short_help="Terminate a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_TERMINATE_EXAMPLE,
)
@click.option("--id", "--job-id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def terminate(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Terminate a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be terminated.
    """
    _validate_job_name_and_id(name=name, id=id)
    anyscale.job.terminate(name=name, id=id, cloud=cloud, project=project)
    if id is not None:
        log.info(f"Query the status of the job with `anyscale job status --id {id}`.")
    else:
        log.info(
            f"Query the status of the job with `anyscale job status --name {name}`."
        )


# TODO(mowen): Add project and cloud support when updating this.
@job_cli.command(
    name="logs", cls=AnyscaleCommand, example=command_examples.JOB_LOGS_EXAMPLE
)
@click.option("--id", "--job-id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option("--run", required=False, help="Name of the job run.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--head",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Used with --max-lines to get `max-lines` lines from the head of the log.",
)
@click.option(
    "--tail",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Used with --max-lines to get `max-lines` lines from the tail of the log.",
)
@click.option(
    "--max-lines",
    required=False,
    default=None,
    type=int,
    help="Used with --head or --tail to limit the number of lines output.",
)
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log.",
)
@click.option(
    "--all-attempts",
    is_flag=True,
    default=False,
    help="DEPRECATED: Listing logs from all attempts no longer supported, instead list jobs by run name.",
)
def logs(  # noqa: PLR0913
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    run: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    head: bool,
    tail: bool,
    max_lines: Optional[int],
    follow: bool = False,
    all_attempts: bool = False,
) -> None:
    """Print the logs of a job.

    By default from the latest job attempt.
    """
    if all_attempts:
        raise click.ClickException(
            "Listing logs from all attempts no longer supported, instead list jobs by run name."
        )
    if follow:
        job_controller = JobController(raise_structured_exception=True)
        job_controller.logs(
            job_id=id, job_name=name, should_follow=True, all_attempts=all_attempts,
        )
    else:
        _validate_job_name_and_id(name=name, id=id)
        if head and tail:
            raise click.ClickException(
                "Only one of '--head' and '--tail' can be provided."
            )
        if max_lines is not None and not (head or tail):
            raise click.ClickException(
                "'--max-lines' must be used with either '--head' or '--tail'"
            )
        mode = JobLogMode.TAIL
        if head:
            mode = JobLogMode.HEAD

        logs = anyscale.job.get_logs(
            id=id,
            name=name,
            run=run,
            cloud=cloud,
            project=project,
            mode=mode,
            max_lines=max_lines,
        )
        print(logs)


@job_cli.command(
    name="wait",
    short_help="Wait for a job to enter a specific state.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_WAIT_EXAMPLE,
)
@click.option(
    "--id", "--job-id", required=False, help="Unique ID of the job.",
)
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--state",
    "-s",
    required=False,
    default=JobState.SUCCEEDED,
    help="The state to wait for this job to enter",
)
@click.option(
    "--timeout-s",
    "--timeout",
    "-t",
    required=False,
    default=1800,
    type=float,
    help="The timeout in seconds after which this command will exit.",
)
def wait(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    state: str = JobState.SUCCEEDED,
    timeout_s=None,
) -> None:
    """Wait for a job to enter a specific state (default: SUCCEEDED).

    To specify the job by name, use the --name flag. To specify the job by id, use the --id flag.

    If the job reaches the target state, the command will exit successfully.

    If the job reaches a terminal state other than the target state, the command will exit with an error.

    If the command reaches the timeout, the command will exit with an error but job execution will continue.
    """
    try:
        state = JobState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))
    try:
        anyscale.job.wait(
            name=name,
            id=id,
            cloud=cloud,
            project=project,
            state=state,
            timeout_s=timeout_s,  # type: ignore
        )
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(str(e)) from None


@job_cli.command(
    name="status",
    short_help="Get the status of a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_STATUS_EXAMPLE,
)
@click.option(
    "--id", "--job-id", required=False, default=None, help="Unique ID of the job."
)
@click.option("--name", "-n", required=False, default=None, help="Name of the job.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the job. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--json",
    "-j",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(
    name: Optional[str],
    id: Optional[str],  # noqa: A002
    cloud: Optional[str],
    project: Optional[str],
    json: bool,
    verbose: bool,
):
    """Query the status of a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be returned.
    """
    _validate_job_name_and_id(name=name, id=id)

    status: JobStatus = anyscale.job.status(
        name=name, id=id, cloud=cloud, project=project
    )
    status_dict = status.to_dict()

    if not verbose:
        status_dict.pop("config", None)

    if json:
        print(json_dumps(status_dict, indent=4, sort_keys=False))
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        print(stream.getvalue(), end="")


@job_cli.group("tags", help="Manage tags for jobs.")
def job_tags_cli() -> None:
    pass


@job_tags_cli.command(
    name="add",
    help="Add or update tags on a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_TAGS_ADD_EXAMPLE,
)
@click.option("--id", "job_id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
def add_tags(job_id: Optional[str], name: Optional[str], tags: Tuple[str]) -> None:
    if not job_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = parse_tags_kv_to_str_map(tags)
    if not tag_map:
        raise click.ClickException("Provide at least one --tag key=value.")
    anyscale.job.add_tags(job_id=job_id, name=name, tags=tag_map)
    stderr = Console(stderr=True)
    ident = job_id or name or "<unknown>"
    stderr.print(f"Tags updated for job '{ident}'.")


@job_tags_cli.command(
    name="remove",
    help="Remove tags by key from a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_TAGS_REMOVE_EXAMPLE,
)
@click.option("--id", "job_id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option("--key", "keys", multiple=True, help="Tag key to remove. Repeatable.")
def remove_tags(job_id: Optional[str], name: Optional[str], keys: Tuple[str]) -> None:
    if not job_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    key_list = [k for k in keys if k and k.strip()]
    if not key_list:
        raise click.ClickException("Provide at least one --key to remove.")
    anyscale.job.remove_tags(job_id=job_id, name=name, keys=key_list)
    stderr = Console(stderr=True)
    ident = job_id or name or "<unknown>"
    stderr.print(f"Removed tag keys {key_list} from job '{ident}'.")


@job_tags_cli.command(
    name="list",
    help="List tags for a job.",
    cls=AnyscaleCommand,
    example=command_examples.JOB_TAGS_LIST_EXAMPLE,
)
@click.option("--id", "job_id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option("--json", "json_output", is_flag=True, default=False, help="JSON output.")
def list_tags(job_id: Optional[str], name: Optional[str], json_output: bool) -> None:
    if not job_id and not name:
        raise click.ClickException("Provide either --id or --name.")
    tag_map = anyscale.job.list_tags(job_id=job_id, name=name)
    if json_output:
        Console().print_json(json=json_dumps(tag_map, indent=2))
    else:
        stderr = Console(stderr=True)
        if not tag_map:
            stderr.print("No tags found.")
            return
        pairs = tag_map.items()
        stderr.print(build_kv_table(pairs, title="Tags"))



import asyncio
from collections import defaultdict
import json
import os
import time
from typing import Dict, List

from anyscale.authenticate import get_auth_api_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.session_command_types import (
    SessionCommandTypes,
)
from anyscale.commands.session_commands_hidden import session_upload_command_logs_impl
from anyscale.webterminal.utils import Command


KILLED_STATUS_CODE = 130
MINUTE_IN_SECONDS = 60
IDLE_TERMINATION_DIR = os.environ.get(
    "IDLE_TERMINATION_DIR", "/tmp/anyscale/idle_termination_reports"
)


class CommandPersister:
    """
        CommandPersister manages persisting the
        state of all commands to our backend.
        It functions as a queue.
    """

    def __init__(self, cli_token: str, host: str, session_id: str):
        self.api_client: DefaultApi = get_auth_api_client(
            cli_token=cli_token, host=host
        ).api_client
        self.cli_token = cli_token
        self.host = host
        self.session_id = session_id
        self.queue: List[Command] = []
        self.running_commands: Dict[str, Command] = {}
        self.last_uploaded_time: Dict[str, float] = defaultdict(float)
        self.is_active = True
        # on first load, report an activity to signal healthy start
        update_idle_termination_status(time.time())

    def shutdown(self) -> None:
        self.is_active = False

    async def persist_commands(self) -> None:  # noqa: PLR0912
        while self.is_active:
            for session_command in self.running_commands.values():
                self.upload_command_log(session_command)
            if len(self.queue) == 0:
                await asyncio.sleep(1)
                continue
            cmd = self.queue.pop(0)
            try:
                if cmd.exec_command:
                    # Mark exec commands as running by default so that we do not try to create this command on the
                    # backend ( it is already created).
                    # When the exec command is finished, command_persister will treat it like other WT commands.
                    self.running_commands[cmd.scid] = cmd
                if cmd.scid not in self.running_commands:
                    if cmd.finished_at:
                        update_idle_termination_status(cmd.finished_at.timestamp())
                    self.api_client.create_session_command_api_v2_sessions_session_id_create_session_command_post(
                        session_id=self.session_id,
                        external_terminal_command={
                            "scid": cmd.scid,
                            "command": cmd.command,
                            "created_at": cmd.created_at,
                            "finished_at": cmd.finished_at,
                            "status_code": cmd.status_code,
                            "web_terminal_tab_id": cmd.term_name,
                            "type": SessionCommandTypes.WEBTERMINAL,
                        },
                    )
                    if not cmd.finished:
                        self.running_commands[cmd.scid] = cmd
                elif cmd.finished:
                    if cmd.finished_at:
                        update_idle_termination_status(cmd.finished_at.timestamp())
                    self.api_client.finish_session_command_api_v2_session_commands_session_command_id_finish_post(
                        session_command_id=cmd.scid,
                        session_command_finish_options={
                            "status_code": cmd.status_code,
                            "stop": False,
                            "terminate": False,
                            "finished_at": cmd.finished_at,
                            "killed_at": cmd.finished_at
                            if cmd.status_code == KILLED_STATUS_CODE
                            else None,
                        },
                    )
                    del self.running_commands[cmd.scid]
                self.upload_command_log(cmd)
                if cmd.finished:
                    del self.last_uploaded_time[cmd.scid]
            except Exception as e:  # noqa: BLE001
                with open(f"/tmp/{cmd.scid}.err", "a") as err_file:
                    err_file.write(f"{e!s}\n")
            if len(self.queue) > 0:
                # If the queue is not empty continue faster through it.
                await asyncio.sleep(0.5)
            else:
                await asyncio.sleep(1)

    def upload_command_log(self, session_command: Command) -> None:
        """
        We upload logs to S3 if any of the below conditions is true:
        1. Session command terminates
        2. Session command runs the first time
        3. The duration from the last uploaded time is larger than the specified period
        """
        current_time_in_s = time.time()
        if (
            session_command.finished
            or current_time_in_s - self.last_uploaded_time[session_command.scid]
            > MINUTE_IN_SECONDS * 2
        ):
            session_upload_command_logs_impl(
                command_id=session_command.scid,
                cli_token=self.cli_token,
                host=self.host,
            )
            # We store the timestamp in which logs have been successfully
            # uploaded to avoid re-uploading large logs that take some time
            # to upload.
            finish_upload_time_in_s = time.time()
            self.last_uploaded_time[session_command.scid] = finish_upload_time_in_s

    def enqueue_command(self, cmd: Command) -> None:
        """
            enqueue_command adds a Command object to be persisted to our backend.
            Only enqueue the command to be peristed if it is
                - just created (therefore not in self.running_commands)
                  we need to create the entry on the backend.
                - finished
                  we need to update the entry on the backend.
        """
        if cmd.scid not in self.running_commands or cmd.finished:
            self.queue.append(cmd)

    def kill_running_commands_for_terminal(self, term_name: str) -> None:
        """
            When we exit a terminal tab, the zsh process is killed and
            all of its children process are also killed. This function then
            enqueues them to be marked as finished.
        """
        commands = list(self.running_commands.values())
        for c in commands:
            if c.term_name == term_name:
                c.finish(KILLED_STATUS_CODE)
                self.enqueue_command(c)


def update_idle_termination_status(timestamp):
    """
        Updates the idle termination status for the web-terminal by storing the last
        activity date as the last finished time of a successful command.
    """
    if not os.path.exists(IDLE_TERMINATION_DIR) or not timestamp:
        return
    with open(f"{IDLE_TERMINATION_DIR}/web_terminal.json", "w") as f:
        f.write(json.dumps({"last_activity_timestamp": timestamp}))



import os
from typing import Any, Dict, List, Optional

import click
import tzlocal
import yaml

from anyscale.anyscale_pydantic import (
    BaseModel,
    Field,
    validator,
)
from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.create_schedule import CreateSchedule
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.schedule_config import (
    ScheduleConfig as APIScheduleConfig,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.models.job_model import JobConfig
from anyscale.project_utils import infer_project_id
from anyscale.tables import SchedulesTable
from anyscale.util import (
    AnyscaleEndpointFormatter,
    populate_unspecified_cluster_configs_from_current_workspace,
    validate_job_config_dict,
)
from anyscale.utils.runtime_env import override_runtime_env_config
from anyscale.utils.workload_types import Workload


log = LogsLogger()


class ScheduleConfig(BaseModel):
    cron_expression: str = Field(
        ...,
        description="A cron expression to define the frequency at which to run this cron job, for example '0 0 * * *' is a cron expression that means 'run at midnight'. Visit crontab.guru to construct a precise cron_expression.",
    )
    timezone: Optional[str] = Field(
        None,
        description="The timezone in which to interpret the cron_expression. Default is Universal time (UTC). In the CLI, you can specify 'local' to use the local timezone",
    )

    @validator("timezone")
    def rewrite_timezone(cls, timezone):
        if timezone == "local":
            local_tz = get_local_timezone()
            log.info(f"Inferred local timezone to be `{local_tz}`")
            return local_tz
        else:
            return timezone


class CreateScheduleConfig(JobConfig):
    schedule: ScheduleConfig


def load_yaml_file_as_dict(config_file: str) -> Dict[str, Any]:
    if not os.path.exists(config_file):
        raise click.ClickException(f"Config file {config_file} not found.")

    with open(config_file) as f:
        config_dict = yaml.safe_load(f)
    return config_dict


def apply_overrides_to_dict(start_dict, **overrides):
    for key, value in overrides.items():
        if value is not None:
            start_dict[key] = value


def load_yaml_file_with_overrides(config_file: str, **overrides):
    d = load_yaml_file_as_dict(config_file)
    apply_overrides_to_dict(d, **overrides)
    return d


def load_schedule_pydantic(config: Dict[str, Any]):
    schedule_config = CreateScheduleConfig.parse_obj(config)
    return schedule_config


def get_local_timezone() -> str:
    return tzlocal.get_localzone_name()


class ScheduleController(BaseController):
    def __init__(
        self, log: Optional[LogsLogger] = None,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__()
        self.log = log
        self.log.open_block("Output")
        self.schedule_api = ScheduleApi(self.api_client)
        self.endpoint_formatter = AnyscaleEndpointFormatter()

    def apply(
        self,
        schedule_config_file: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> str:
        schedule_config = self._resolve_config(
            schedule_config_file, name=name, description=description
        )
        id = self.schedule_api.apply(schedule_config)  # noqa: A001
        self.url(id)
        self.log.info(f"Schedule {id} has been updated successfully.")
        return id

    def _resolve_config(
        self, schedule_config_file: str, **overrides
    ) -> CreateScheduleConfig:
        schedule_config_dict = load_yaml_file_with_overrides(
            schedule_config_file, **overrides
        )
        validate_job_config_dict(schedule_config_dict, self.api_client)

        # If running in a workspace, auto-populate unspecified fields.
        schedule_config_dict = populate_unspecified_cluster_configs_from_current_workspace(
            schedule_config_dict, self.anyscale_api_client,
        )

        schedule_config: CreateScheduleConfig = CreateScheduleConfig.parse_obj(
            schedule_config_dict
        )

        schedule_config.project_id = infer_project_id(
            self.anyscale_api_client,
            self.api_client,
            self.log,
            project_id=schedule_config.project_id,
            cluster_compute_id=schedule_config.compute_config_id,
            cluster_compute=schedule_config.compute_config,
            cloud=schedule_config.cloud,
        )
        schedule_config.runtime_env = override_runtime_env_config(
            runtime_env=schedule_config.runtime_env,
            anyscale_api_client=self.anyscale_api_client,
            api_client=self.api_client,
            workload_type=Workload.SCHEDULED_JOBS,
            compute_config_id=schedule_config.compute_config_id,
            log=self.log,
        )

        return schedule_config

    def resolve_file_name_or_id(
        self,
        *,
        schedule_config_file: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
    ) -> str:
        """Try to resolve a Schedule Config File, Name, or Id to a Schedule Id

        At least one of the inputs must be non null.

        Args:
            schedule_config_file (Optional[str]): A file name
            id (Optional[str]): An id
            name (Optional[str]): a name

        Raises:
            click.ClickException: If the id could not be uniquely determined.

        Returns:
            str: the id of the schedule
        """

        assert_single_parameter(
            schedule_config_file=schedule_config_file, id=id, name=name
        )

        if id:
            try:
                result = self.schedule_api.get(id)
                return result.id
            except Exception as e:  # noqa: BLE001
                raise click.ClickException(f"No schedule found with id `{id}`") from e
        elif name:
            schedules = self.schedule_api.list(name=name)
            schedules = [schedule for schedule in schedules if schedule.name == name]

            if len(schedules) == 1:
                return schedules[0].id
            elif len(schedules) > 1:
                self.log.error(str(SchedulesTable(schedules)))
                raise click.ClickException(
                    f"Found multiple schedules matching the name {name}"
                )
            else:
                raise click.ClickException(
                    f"Found no schedules matching the name {name}"
                )
        elif schedule_config_file:
            config = self._resolve_config(schedule_config_file)
            name = config.name
            project_id = config.project_id

            # Schedules are unique by name and project_id
            schedules = self.schedule_api.list(name=name, project_id=project_id)
            if len(schedules) == 1:
                return schedules[0].id
            elif len(schedules) > 1:
                self.log.error(str(SchedulesTable(schedules)))
                raise click.ClickException(
                    f"Unable to determine which of the following schedules this file `{schedule_config_file}` is referring to. Please make sure `name` and `project_id` are specified in your schedule yaml."
                )
            else:
                raise click.ClickException(
                    f"Found no schedules matching the file `{schedule_config_file}"
                )

        raise click.ClickException(
            "Please specify a file, a name, or an id to resolve your schedule"
        )

    def list(self, project_id=None, name=None, creator_id=None, id=None):  # noqa: A002
        if id:
            if any([project_id, name, creator_id]):
                raise click.ClickException(
                    "If `id` is specified, no other filters can be specified."
                )
            schedules = [self.schedule_api.get(id)]
        else:
            schedules = self.schedule_api.list(
                project_id=project_id, name=name, creator_id=creator_id
            )
        print(SchedulesTable(schedules))

    def pause(self, id, is_paused=True):  # noqa: A002
        schedule = self.schedule_api.pause(id, is_paused)
        text = "paused" if is_paused else "resumed"
        self.url(id)
        self.log.info(f"Scheduled Job {id} has been {text}.")
        if not is_paused:
            self.log.info(f"Next trigger at {schedule.next_trigger_at}")

    def trigger(self, id):  # noqa: A002
        created_job = self.schedule_api.trigger(id)
        self.url(id)
        self.log.info(f"Successfully created Job with id {created_job.id}")
        url = self.endpoint_formatter.get_job_endpoint(created_job.id)
        self.log.info(f"View your Production Job at {url}")

    def url(self, id) -> str:  # noqa: A002
        url = self.endpoint_formatter.get_schedule_endpoint(id)
        self.log.info(f"View your schedule at {url}")
        return url


class ScheduleApi:
    def __init__(self, api_client: DefaultApi = None):
        self.api_client = api_client or get_auth_api_client().api_client

    def apply(self, schedule_config: CreateScheduleConfig) -> str:
        formatted = format_create_schedule_config(schedule_config)
        job = self.api_client.create_or_update_job_api_v2_experimental_cron_jobs_put(
            formatted
        ).result
        return job.id

    def list(self, project_id=None, creator_id=None, name=None) -> List[Any]:
        return self.api_client.list_cron_jobs_api_v2_experimental_cron_jobs_get(
            project_id=project_id, creator_id=creator_id, name=name
        ).results

    def get(self, id: str) -> Any:  # noqa: A002
        return self.api_client.get_cron_job_api_v2_experimental_cron_jobs_cron_job_id_get(
            id
        ).result

    def pause(self, cron_job_id: str, is_paused=True):
        return self.api_client.pause_cron_job_api_v2_experimental_cron_jobs_cron_job_id_pause_post(
            cron_job_id, {"is_paused": is_paused}
        ).result

    def trigger(self, cron_job_id: str):
        return self.api_client.trigger_cron_job_api_v2_experimental_cron_jobs_cron_job_id_trigger_post(
            cron_job_id
        ).result

    def get_current_user(self) -> str:
        user = self.api_client.get_user_info_api_v2_userinfo_get().result
        return user.id


def format_schedule_config(schedule: ScheduleConfig):
    return APIScheduleConfig(
        cron_expression=schedule.cron_expression, timezone=schedule.timezone,
    )


def format_api_job_config(job_config: JobConfig):
    return ProductionJobConfig(
        entrypoint=job_config.entrypoint,
        runtime_env=job_config.runtime_env,
        build_id=job_config.build_id,
        compute_config_id=job_config.compute_config_id,
        max_retries=job_config.max_retries,
    )


def format_create_schedule_config(schedule_config: CreateScheduleConfig):
    return CreateSchedule(
        name=schedule_config.name,
        description=schedule_config.description,
        project_id=schedule_config.project_id,
        config=format_api_job_config(schedule_config),
        schedule=format_schedule_config(schedule_config.schedule),
    )


def assert_single_parameter(**kwargs) -> None:
    keys_not_none = [a for a in kwargs if kwargs[a] is not None]
    if len(keys_not_none) == 0:
        raise click.ClickException(
            f"At least one of the following flags must be set: {kwargs.keys()}"
        )
    if len(keys_not_none) > 1:
        raise click.ClickException(
            f"Only one of the following flags can be set: {keys_not_none}"
        )



import contextlib
from enum import Enum
import sys
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar
import uuid

import boto3
from botocore.exceptions import ClientError
from click.exceptions import ClickException

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models import (
    CloudAnalyticsEvent,
    CloudAnalyticsEventCloudResource,
    CloudAnalyticsEventCommandName,
    CloudAnalyticsEventError,
    CloudAnalyticsEventName,
    CloudProviders,
    CreateAnalyticsEvent,
)
from anyscale.cloud_utils import get_cloud_id_and_name
from anyscale.shared_anyscale_utils.utils.collections import flatten


V = TypeVar("V")


def get_organization_default_cloud(api_client: DefaultApi) -> Optional[str]:
    """Return default cloud name for organization if it exists and
        if user has correct permissions for it.

        Returns:
            Name of default cloud name for organization if it exists and
            if user has correct permissions for it.
        """
    user = api_client.get_user_info_api_v2_userinfo_get().result
    organization = user.organizations[0]  # Each user only has one org
    if organization.default_cloud_id:
        try:
            # Check permissions
            _, cloud_name = get_cloud_id_and_name(
                api_client, cloud_id=organization.default_cloud_id
            )
            return str(cloud_name)
        except Exception:  # noqa: BLE001
            return None
    return None


def get_default_cloud(
    api_client: DefaultApi, cloud_name: Optional[str]
) -> Tuple[str, str]:
    """Returns the cloud id from cloud name.
    If cloud name is not provided, returns the default cloud name if exists in organization.
    If default cloud name does not exist returns last used cloud.
    """

    if cloud_name is None:
        default_cloud_name = get_organization_default_cloud(api_client)
        if default_cloud_name:
            cloud_name = default_cloud_name
    return get_cloud_id_and_name(api_client, cloud_name=cloud_name)


def verify_anyscale_access(
    api_client: DefaultApi,
    cloud_id: str,
    cloud_provider: CloudProviders,
    logger: CloudSetupLogger,
) -> bool:
    try:
        api_client.verify_access_api_v2_cloudsverify_access_cloud_id_get(cloud_id)
        return True
    except ClickException as e:
        if cloud_provider == CloudProviders.AWS:
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.AWS_IAM_ROLE,
                CloudSetupError.ANYSCALE_ACCESS_DENIED,
            )
        elif cloud_provider == CloudProviders.GCP:
            logger.log_resource_error(
                CloudAnalyticsEventCloudResource.GCP_SERVICE_ACCOUNT,
                CloudSetupError.ANYSCALE_ACCESS_DENIED,
            )
        logger.error(
            f"Anyscale's control plane is unable to access resources on your cloud provider.\n{e}"
        )
        return False


def modify_memorydb_parameter_group(
    parameter_group_name: str,
    region: str,
    boto3_session: Optional[boto3.Session] = None,
) -> None:
    """
    Modify the memorydb parameter group to set the maxmemory-policy to allkeys-lru.

    This is not done in the cloudformation template because we have to create the paramter group first and then modify it.
    """
    try:
        if boto3_session is None:
            boto3_session = boto3.Session()
        memorydb_client = boto3_session.client("memorydb", region_name=region)
        memorydb_client.update_parameter_group(
            ParameterGroupName=parameter_group_name,
            ParameterNameValues=[
                {"ParameterName": "maxmemory-policy", "ParameterValue": "allkeys-lru",}
            ],
        )
    except ClientError as e:
        # TODO (allenyin): add memorydb to the cloud provider error list.
        raise ClickException(
            f"Failed to modify memorydb parameter group {parameter_group_name}. Please make sure you have permission to perform UpdateParameterGroup on memorydb clusters and try again. \n{e}"
        )


def wait_for_lb_resource_termination(
    api_client: DefaultApi,
    cloud_id: str,
    timeout_s: int = 900,  # 15 minute timeout
    poll_interval_s: int = 10,  # Poll every 10 seconds
):
    start = time.time()
    while time.time() - start < timeout_s:
        response = api_client.get_lb_resource_api_v2_clouds_cloud_id_get_lb_resource_post(
            cloud_id=cloud_id
        ).result
        if response.is_terminated:
            break
        time.sleep(poll_interval_s)
    else:
        raise ClickException(
            f"LB resources and namespace termination timed out after {timeout_s} seconds."
        )


class CloudSetupError(str, Enum):
    ANYSCALE_ACCESS_DENIED = "ANYSCALE_ACCESS_DENIED"
    RESOURCE_NOT_FOUND = "RESOURCE_NOT_FOUND"
    CIDR_BLOCK_TOO_SMALL = "CIDR_BLOCK_TOO_SMALL"
    MISSING_CLOUD_RESOURCE_ID = "MISSING_CLOUD_RESOURCE_ID"
    ONLY_ONE_SUBNET = "ONLY_ONE_SUBNET"
    SUBNET_NOT_IN_VPC = "SUBNET_NOT_IN_VPC"
    ONLY_ONE_AZ = "ONLY_ONE_AZ"
    IAM_ROLE_ACCOUNT_MISMATCH = "IAM_ROLE_ACCOUNT_MISMATCH"
    INSTANCE_PROFILE_NOT_FOUND = "INSTANCE_PROFILE_NOT_FOUND"
    INTERNAL_COMMUNICATION_NOT_ALLOWED = "INTERNAL_COMMUNICATION_NOT_ALLOWED"
    MALFORMED_CORS_RULE = "MALFORMED_CORS_RULE"
    INCORRECT_CORS_RULE = "INCORRECT_CORS_RULE"
    MOUNT_TARGET_NOT_FOUND = "MOUNT_TARGET_NOT_FOUND"
    INVALID_MOUNT_TARGET = "INVALID_MOUNT_TARGET"
    PROJECT_NOT_ACTIVE = "PROJECT_NOT_ACTIVE"
    API_NOT_ENABLED = "API_NOT_ENABLED"
    FIREWALL_NOT_ASSOCIATED_WITH_VPC = "FIREWALL_NOT_ASSOCIATED_WITH_VPC"
    FILESTORE_NAME_MALFORMED = "FILESTORE_NAME_MALFORMED"
    FILESTORE_NOT_CONNECTED_TO_VPC = "FILESTORE_NOT_CONNECTED_TO_VPC"
    MEMORYSTORE_NOT_CONNECTED_TO_VPC = "MEMORYSTORE_NOT_CONNECTED_TO_VPC"
    MEMORYSTORE_READ_REPLICAS_DISABLED = "MEMORYSTORE_READ_REPLICAS_DISABLED"
    MEMORYDB_CLUSTER_UNAVAILABLE = "MEMORYDB_CLUSTER_UNAVAILABLE"


class CloudEventProducer:
    """
    Produce events during cloud setup/register/verify
    """

    def __init__(self, cli_version: str, api_client: DefaultApi):
        self.api_client = api_client
        self.cloud_id: Optional[str] = None
        self.cloud_provider: Optional[CloudProviders] = None
        self.cli_version = cli_version

    def init_trace_context(
        self,
        command_name: CloudAnalyticsEventCommandName,
        cloud_provider: CloudProviders,
        cloud_id: Optional[str] = None,
    ):
        self.trace_id = str(uuid.uuid4().hex)
        self.command_name = command_name
        self.raw_command_input = str(" ".join(sys.argv[1:]))
        self.cloud_id = cloud_id
        self.cloud_provider = cloud_provider

    def set_cloud_id(self, cloud_id: str):
        self.cloud_id = cloud_id

    def produce(
        self,
        event_name: CloudAnalyticsEventName,
        succeeded: bool,
        logger: Optional[CloudSetupLogger] = None,
        internal_error: Optional[str] = None,
    ):
        with contextlib.suppress(Exception):
            # shouldn't block cloud setup even if cloud event generation fails
            error = None
            if not succeeded:
                cloud_provider_error = None
                if logger:
                    cloud_provider_error_list = logger.get_cloud_provider_errors()
                    logger.clear_cloud_provider_errors()
                    if len(cloud_provider_error_list) > 0:
                        cloud_provider_error = cloud_provider_error_list
                error = CloudAnalyticsEventError(
                    internal_error=internal_error,
                    cloud_provider_error=cloud_provider_error,
                )

            self.api_client.produce_analytics_event_api_v2_analytics_post(
                CreateAnalyticsEvent(
                    cloud_analytics_event=CloudAnalyticsEvent(
                        cli_version=self.cli_version,
                        trace_id=self.trace_id,
                        cloud_id=self.cloud_id,
                        succeeded=succeeded,
                        command_name=self.command_name,
                        raw_command_input=self.raw_command_input,
                        cloud_provider=self.cloud_provider,
                        event_name=event_name,
                        error=error,
                    )
                )
            )


def validate_aws_credentials(
    logger: CloudSetupLogger, boto3_session: Optional[Any] = None
) -> bool:
    try:
        if boto3_session is None:
            boto3_session = boto3.Session()
        boto3.client("sts").get_caller_identity()
        return True
    except Exception as e:  # noqa: BLE001
        logger.error(f"Failed to validate AWS credentials: {e}")
        return False


def get_errored_resources_and_reasons(
    cfn_client: Any, stack_name: str
) -> Dict[str, str]:
    """
    Describes the CloudFormation stack events and extracts the failure reasons.
    """
    error_details: Dict[str, str] = {}
    response = cfn_client.describe_stack_events(StackName=stack_name)
    error_details.update(
        extract_cloudformation_failure_reasons(response["StackEvents"])
    )
    while response.get("NextToken") is not None:
        response = cfn_client.describe_stack_events(
            StackName=stack_name, NextToken=response["NextToken"]
        )
        error_details.update(
            extract_cloudformation_failure_reasons(response["StackEvents"])
        )
    return error_details


def extract_cloudformation_failure_reasons(
    events: List[Dict[str, Any]]
) -> Dict[str, str]:
    """
    Extracts the failure reasons from the CloudFormation events.
    """
    failure_reasons: Dict[str, str] = {}
    for event in events:
        if event.get("ResourceStatus") == "CREATE_FAILED":
            error_resource = event.get("LogicalResourceId", "UnknownResource")
            error_reason = event.get("ResourceStatusReason", "Unknown reason")
            if "Resource creation cancelled" in error_reason:
                # Resource creation cancelled due to other resource failures.
                # Skip this error.
                continue
            failure_reasons[error_resource] = error_reason
    return failure_reasons


def _unroll_resources_for_aws_list_call(aws_list_function: Callable, list_key: str):
    def token_extractor(response) -> Optional[str]:
        return response.get("NextToken")

    # NOTE: `NextToken` is expected to have `str` type and therefore we have
    #       we have to workaround its optionality by completely omitting it
    #       from the call
    responses = unroll_pagination(
        lambda next_token: aws_list_function(
            **({"NextToken": next_token} if next_token else {}),  # type: ignore
        ),
        token_extractor,
    )

    return flatten(*[r[list_key] for r in responses if list_key in r])


def unroll_pagination(
    paginated_method: Callable, next_token_extractor: Callable,
):
    """Handles paginated method's invocation by
    - Repeatedly invoking the method (injecting the token from its previous response)
    - Collecting all of the responses in a list
    """

    next_token: Optional[str] = None
    results = []
    while True:
        r = paginated_method(next_token)
        results.append(r)
        next_token = next_token_extractor(r)
        if not next_token:
            break

    return results



import click


# TODO: type the return value
def try_import_gcp_secretmanager():
    try:
        from google.cloud import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for secrets manager")
            secretmanager,
        )

        return secretmanager
    except ImportError:
        raise click.ClickException(
            "pip package `google-cloud-secret-manager` is not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_utils():
    try:
        from anyscale.utils import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for CLI commands")
            gcp_utils,
        )

        return gcp_utils
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_verify_lib():
    try:
        import anyscale.gcp_verification as verify_lib  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for CLI commands")

        return verify_lib
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_managed_setup_utils():
    try:
        from anyscale.utils import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for CLI commands")
            gcp_managed_setup_utils,
        )

        return gcp_managed_setup_utils
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_exceptions():
    try:
        from google.api_core.exceptions import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for CLI commands")
            GoogleAPICallError,
        )
        from googleapiclient.errors import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP dependency for CLI commands")
            HttpError,
        )

        return GoogleAPICallError, HttpError
    except ImportError as e:
        raise click.ClickException(
            f"Import error occurred: {e}. Please install GCP-related packages with `pip install 'anyscale[gcp]'`."
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CloudDataBucketPresignedUrlResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'file_uri': 'str',
        'url': 'str',
        'file_exists': 'bool',
        'bucket_name': 'str',
        'bucket_path': 'str',
        'url_scheme': 'CloudDataBucketPresignedUrlScheme',
        'file_size': 'int'
    }

    attribute_map = {
        'file_uri': 'file_uri',
        'url': 'url',
        'file_exists': 'file_exists',
        'bucket_name': 'bucket_name',
        'bucket_path': 'bucket_path',
        'url_scheme': 'url_scheme',
        'file_size': 'file_size'
    }

    def __init__(self, file_uri=None, url=None, file_exists=None, bucket_name=None, bucket_path=None, url_scheme=None, file_size=0, local_vars_configuration=None):  # noqa: E501
        """CloudDataBucketPresignedUrlResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._file_uri = None
        self._url = None
        self._file_exists = None
        self._bucket_name = None
        self._bucket_path = None
        self._url_scheme = None
        self._file_size = None
        self.discriminator = None

        self.file_uri = file_uri
        self.url = url
        self.file_exists = file_exists
        self.bucket_name = bucket_name
        self.bucket_path = bucket_path
        self.url_scheme = url_scheme
        if file_size is not None:
            self.file_size = file_size

    @property
    def file_uri(self):
        """Gets the file_uri of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        The fully-qualified path to the file in the cloud data bucket (e.g. 's3://anyscale-production-data-cld-123/org_id/cloud_id/runtime_env_packages/job_id.zip').  # noqa: E501

        :return: The file_uri of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: str
        """
        return self._file_uri

    @file_uri.setter
    def file_uri(self, file_uri):
        """Sets the file_uri of this CloudDataBucketPresignedUrlResponse.

        The fully-qualified path to the file in the cloud data bucket (e.g. 's3://anyscale-production-data-cld-123/org_id/cloud_id/runtime_env_packages/job_id.zip').  # noqa: E501

        :param file_uri: The file_uri of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and file_uri is None:  # noqa: E501
            raise ValueError("Invalid value for `file_uri`, must not be `None`")  # noqa: E501

        self._file_uri = file_uri

    @property
    def url(self):
        """Gets the url of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        The presigned URL to use for uploading or downloading the file in the cloud data bucket.  # noqa: E501

        :return: The url of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this CloudDataBucketPresignedUrlResponse.

        The presigned URL to use for uploading or downloading the file in the cloud data bucket.  # noqa: E501

        :param url: The url of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and url is None:  # noqa: E501
            raise ValueError("Invalid value for `url`, must not be `None`")  # noqa: E501

        self._url = url

    @property
    def file_exists(self):
        """Gets the file_exists of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        Whether or not an object under the returned URI already exists.  # noqa: E501

        :return: The file_exists of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: bool
        """
        return self._file_exists

    @file_exists.setter
    def file_exists(self, file_exists):
        """Sets the file_exists of this CloudDataBucketPresignedUrlResponse.

        Whether or not an object under the returned URI already exists.  # noqa: E501

        :param file_exists: The file_exists of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and file_exists is None:  # noqa: E501
            raise ValueError("Invalid value for `file_exists`, must not be `None`")  # noqa: E501

        self._file_exists = file_exists

    @property
    def bucket_name(self):
        """Gets the bucket_name of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        The name of the cloud data bucket where the file will be uploaded (e.g. 'anyscale-production-data-cld-123').  # noqa: E501

        :return: The bucket_name of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: str
        """
        return self._bucket_name

    @bucket_name.setter
    def bucket_name(self, bucket_name):
        """Sets the bucket_name of this CloudDataBucketPresignedUrlResponse.

        The name of the cloud data bucket where the file will be uploaded (e.g. 'anyscale-production-data-cld-123').  # noqa: E501

        :param bucket_name: The bucket_name of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and bucket_name is None:  # noqa: E501
            raise ValueError("Invalid value for `bucket_name`, must not be `None`")  # noqa: E501

        self._bucket_name = bucket_name

    @property
    def bucket_path(self):
        """Gets the bucket_path of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        The path to the file in the cloud data bucket (e.g. 'org_id/cloud_id/runtime_env_packages/job_id.zip').  # noqa: E501

        :return: The bucket_path of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: str
        """
        return self._bucket_path

    @bucket_path.setter
    def bucket_path(self, bucket_path):
        """Sets the bucket_path of this CloudDataBucketPresignedUrlResponse.

        The path to the file in the cloud data bucket (e.g. 'org_id/cloud_id/runtime_env_packages/job_id.zip').  # noqa: E501

        :param bucket_path: The bucket_path of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and bucket_path is None:  # noqa: E501
            raise ValueError("Invalid value for `bucket_path`, must not be `None`")  # noqa: E501

        self._bucket_path = bucket_path

    @property
    def url_scheme(self):
        """Gets the url_scheme of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        The scheme used to upload or download files using the presigned URL.  # noqa: E501

        :return: The url_scheme of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: CloudDataBucketPresignedUrlScheme
        """
        return self._url_scheme

    @url_scheme.setter
    def url_scheme(self, url_scheme):
        """Sets the url_scheme of this CloudDataBucketPresignedUrlResponse.

        The scheme used to upload or download files using the presigned URL.  # noqa: E501

        :param url_scheme: The url_scheme of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: CloudDataBucketPresignedUrlScheme
        """
        if self.local_vars_configuration.client_side_validation and url_scheme is None:  # noqa: E501
            raise ValueError("Invalid value for `url_scheme`, must not be `None`")  # noqa: E501

        self._url_scheme = url_scheme

    @property
    def file_size(self):
        """Gets the file_size of this CloudDataBucketPresignedUrlResponse.  # noqa: E501

        For READ requests, the size of the file in bytes, if it exists.  # noqa: E501

        :return: The file_size of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :rtype: int
        """
        return self._file_size

    @file_size.setter
    def file_size(self, file_size):
        """Sets the file_size of this CloudDataBucketPresignedUrlResponse.

        For READ requests, the size of the file in bytes, if it exists.  # noqa: E501

        :param file_size: The file_size of this CloudDataBucketPresignedUrlResponse.  # noqa: E501
        :type: int
        """

        self._file_size = file_size

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CloudDataBucketPresignedUrlResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CloudDataBucketPresignedUrlResponse):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class RayRuntimeEnvConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'working_dir': 'str',
        'py_modules': 'list[str]',
        'relative_working_dir': 'str',
        'relative_py_modules': 'list[str]',
        'py_executable': 'str',
        'pip': 'list[str]',
        'conda': 'object',
        'env_vars': 'dict(str, str)',
        'config': 'object',
        'image_uri': 'str'
    }

    attribute_map = {
        'working_dir': 'working_dir',
        'py_modules': 'py_modules',
        'relative_working_dir': 'relative_working_dir',
        'relative_py_modules': 'relative_py_modules',
        'py_executable': 'py_executable',
        'pip': 'pip',
        'conda': 'conda',
        'env_vars': 'env_vars',
        'config': 'config',
        'image_uri': 'image_uri'
    }

    def __init__(self, working_dir=None, py_modules=None, relative_working_dir=None, relative_py_modules=None, py_executable=None, pip=None, conda=None, env_vars=None, config=None, image_uri=None, local_vars_configuration=None):  # noqa: E501
        """RayRuntimeEnvConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._working_dir = None
        self._py_modules = None
        self._relative_working_dir = None
        self._relative_py_modules = None
        self._py_executable = None
        self._pip = None
        self._conda = None
        self._env_vars = None
        self._config = None
        self._image_uri = None
        self.discriminator = None

        if working_dir is not None:
            self.working_dir = working_dir
        if py_modules is not None:
            self.py_modules = py_modules
        if relative_working_dir is not None:
            self.relative_working_dir = relative_working_dir
        if relative_py_modules is not None:
            self.relative_py_modules = relative_py_modules
        if py_executable is not None:
            self.py_executable = py_executable
        if pip is not None:
            self.pip = pip
        if conda is not None:
            self.conda = conda
        if env_vars is not None:
            self.env_vars = env_vars
        if config is not None:
            self.config = config
        if image_uri is not None:
            self.image_uri = image_uri

    @property
    def working_dir(self):
        """Gets the working_dir of this RayRuntimeEnvConfig.  # noqa: E501

        The working directory that your code will run in. Must be a remote URI like an s3 or git path.  # noqa: E501

        :return: The working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: str
        """
        return self._working_dir

    @working_dir.setter
    def working_dir(self, working_dir):
        """Sets the working_dir of this RayRuntimeEnvConfig.

        The working directory that your code will run in. Must be a remote URI like an s3 or git path.  # noqa: E501

        :param working_dir: The working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :type: str
        """

        self._working_dir = working_dir

    @property
    def py_modules(self):
        """Gets the py_modules of this RayRuntimeEnvConfig.  # noqa: E501

        Python modules that will be installed along with your runtime env. These must be remote URIs.  # noqa: E501

        :return: The py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._py_modules

    @py_modules.setter
    def py_modules(self, py_modules):
        """Sets the py_modules of this RayRuntimeEnvConfig.

        Python modules that will be installed along with your runtime env. These must be remote URIs.  # noqa: E501

        :param py_modules: The py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :type: list[str]
        """

        self._py_modules = py_modules

    @property
    def relative_working_dir(self):
        """Gets the relative_working_dir of this RayRuntimeEnvConfig.  # noqa: E501

        Relative path to the working directory that your code will run in. The appropriate cloud deployment object storage will be prepended to this path.  # noqa: E501

        :return: The relative_working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: str
        """
        return self._relative_working_dir

    @relative_working_dir.setter
    def relative_working_dir(self, relative_working_dir):
        """Sets the relative_working_dir of this RayRuntimeEnvConfig.

        Relative path to the working directory that your code will run in. The appropriate cloud deployment object storage will be prepended to this path.  # noqa: E501

        :param relative_working_dir: The relative_working_dir of this RayRuntimeEnvConfig.  # noqa: E501
        :type: str
        """

        self._relative_working_dir = relative_working_dir

    @property
    def relative_py_modules(self):
        """Gets the relative_py_modules of this RayRuntimeEnvConfig.  # noqa: E501

        Relative paths to python modules that will be installed along with your runtime env. The appropriate cloud deployment object storage will be prepended to these paths. If `py_modules` are specified, they will be also be installed.  # noqa: E501

        :return: The relative_py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._relative_py_modules

    @relative_py_modules.setter
    def relative_py_modules(self, relative_py_modules):
        """Sets the relative_py_modules of this RayRuntimeEnvConfig.

        Relative paths to python modules that will be installed along with your runtime env. The appropriate cloud deployment object storage will be prepended to these paths. If `py_modules` are specified, they will be also be installed.  # noqa: E501

        :param relative_py_modules: The relative_py_modules of this RayRuntimeEnvConfig.  # noqa: E501
        :type: list[str]
        """

        self._relative_py_modules = relative_py_modules

    @property
    def py_executable(self):
        """Gets the py_executable of this RayRuntimeEnvConfig.  # noqa: E501

        Specifies the executable used for running the Ray workers. It can include arguments as well.  # noqa: E501

        :return: The py_executable of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: str
        """
        return self._py_executable

    @py_executable.setter
    def py_executable(self, py_executable):
        """Sets the py_executable of this RayRuntimeEnvConfig.

        Specifies the executable used for running the Ray workers. It can include arguments as well.  # noqa: E501

        :param py_executable: The py_executable of this RayRuntimeEnvConfig.  # noqa: E501
        :type: str
        """

        self._py_executable = py_executable

    @property
    def pip(self):
        """Gets the pip of this RayRuntimeEnvConfig.  # noqa: E501

        A list of pip packages to install.  # noqa: E501

        :return: The pip of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._pip

    @pip.setter
    def pip(self, pip):
        """Sets the pip of this RayRuntimeEnvConfig.

        A list of pip packages to install.  # noqa: E501

        :param pip: The pip of this RayRuntimeEnvConfig.  # noqa: E501
        :type: list[str]
        """

        self._pip = pip

    @property
    def conda(self):
        """Gets the conda of this RayRuntimeEnvConfig.  # noqa: E501

        [Union[Dict[str, Any], str]: Either the conda YAML config or the name of a local conda env (e.g., \"pytorch_p36\"),   # noqa: E501

        :return: The conda of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: object
        """
        return self._conda

    @conda.setter
    def conda(self, conda):
        """Sets the conda of this RayRuntimeEnvConfig.

        [Union[Dict[str, Any], str]: Either the conda YAML config or the name of a local conda env (e.g., \"pytorch_p36\"),   # noqa: E501

        :param conda: The conda of this RayRuntimeEnvConfig.  # noqa: E501
        :type: object
        """

        self._conda = conda

    @property
    def env_vars(self):
        """Gets the env_vars of this RayRuntimeEnvConfig.  # noqa: E501

        Environment variables to set.  # noqa: E501

        :return: The env_vars of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this RayRuntimeEnvConfig.

        Environment variables to set.  # noqa: E501

        :param env_vars: The env_vars of this RayRuntimeEnvConfig.  # noqa: E501
        :type: dict(str, str)
        """

        self._env_vars = env_vars

    @property
    def config(self):
        """Gets the config of this RayRuntimeEnvConfig.  # noqa: E501

        Config for runtime environment. Can be used to setup setup_timeout_seconds, the timeout of runtime environment creation.  # noqa: E501

        :return: The config of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: object
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this RayRuntimeEnvConfig.

        Config for runtime environment. Can be used to setup setup_timeout_seconds, the timeout of runtime environment creation.  # noqa: E501

        :param config: The config of this RayRuntimeEnvConfig.  # noqa: E501
        :type: object
        """

        self._config = config

    @property
    def image_uri(self):
        """Gets the image_uri of this RayRuntimeEnvConfig.  # noqa: E501

        Specifies the image URI of the container in which the job will run.  # noqa: E501

        :return: The image_uri of this RayRuntimeEnvConfig.  # noqa: E501
        :rtype: str
        """
        return self._image_uri

    @image_uri.setter
    def image_uri(self, image_uri):
        """Sets the image_uri of this RayRuntimeEnvConfig.

        Specifies the image URI of the container in which the job will run.  # noqa: E501

        :param image_uri: The image_uri of this RayRuntimeEnvConfig.  # noqa: E501
        :type: str
        """

        self._image_uri = image_uri

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, RayRuntimeEnvConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, RayRuntimeEnvConfig):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SCIMMigrationResult(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'organization_id': 'str',
        'total_users_processed': 'int',
        'users_migrated': 'list[UserMigrationResult]',
        'errors': 'list[str]',
        'users_to_remove': 'list[UserToRemove]',
        'dry_run': 'bool'
    }

    attribute_map = {
        'organization_id': 'organization_id',
        'total_users_processed': 'total_users_processed',
        'users_migrated': 'users_migrated',
        'errors': 'errors',
        'users_to_remove': 'users_to_remove',
        'dry_run': 'dry_run'
    }

    def __init__(self, organization_id=None, total_users_processed=None, users_migrated=None, errors=None, users_to_remove=None, dry_run=True, local_vars_configuration=None):  # noqa: E501
        """SCIMMigrationResult - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._organization_id = None
        self._total_users_processed = None
        self._users_migrated = None
        self._errors = None
        self._users_to_remove = None
        self._dry_run = None
        self.discriminator = None

        self.organization_id = organization_id
        self.total_users_processed = total_users_processed
        self.users_migrated = users_migrated
        self.errors = errors
        if users_to_remove is not None:
            self.users_to_remove = users_to_remove
        if dry_run is not None:
            self.dry_run = dry_run

    @property
    def organization_id(self):
        """Gets the organization_id of this SCIMMigrationResult.  # noqa: E501


        :return: The organization_id of this SCIMMigrationResult.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this SCIMMigrationResult.


        :param organization_id: The organization_id of this SCIMMigrationResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_id is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def total_users_processed(self):
        """Gets the total_users_processed of this SCIMMigrationResult.  # noqa: E501


        :return: The total_users_processed of this SCIMMigrationResult.  # noqa: E501
        :rtype: int
        """
        return self._total_users_processed

    @total_users_processed.setter
    def total_users_processed(self, total_users_processed):
        """Sets the total_users_processed of this SCIMMigrationResult.


        :param total_users_processed: The total_users_processed of this SCIMMigrationResult.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and total_users_processed is None:  # noqa: E501
            raise ValueError("Invalid value for `total_users_processed`, must not be `None`")  # noqa: E501

        self._total_users_processed = total_users_processed

    @property
    def users_migrated(self):
        """Gets the users_migrated of this SCIMMigrationResult.  # noqa: E501


        :return: The users_migrated of this SCIMMigrationResult.  # noqa: E501
        :rtype: list[UserMigrationResult]
        """
        return self._users_migrated

    @users_migrated.setter
    def users_migrated(self, users_migrated):
        """Sets the users_migrated of this SCIMMigrationResult.


        :param users_migrated: The users_migrated of this SCIMMigrationResult.  # noqa: E501
        :type: list[UserMigrationResult]
        """
        if self.local_vars_configuration.client_side_validation and users_migrated is None:  # noqa: E501
            raise ValueError("Invalid value for `users_migrated`, must not be `None`")  # noqa: E501

        self._users_migrated = users_migrated

    @property
    def errors(self):
        """Gets the errors of this SCIMMigrationResult.  # noqa: E501


        :return: The errors of this SCIMMigrationResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._errors

    @errors.setter
    def errors(self, errors):
        """Sets the errors of this SCIMMigrationResult.


        :param errors: The errors of this SCIMMigrationResult.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and errors is None:  # noqa: E501
            raise ValueError("Invalid value for `errors`, must not be `None`")  # noqa: E501

        self._errors = errors

    @property
    def users_to_remove(self):
        """Gets the users_to_remove of this SCIMMigrationResult.  # noqa: E501


        :return: The users_to_remove of this SCIMMigrationResult.  # noqa: E501
        :rtype: list[UserToRemove]
        """
        return self._users_to_remove

    @users_to_remove.setter
    def users_to_remove(self, users_to_remove):
        """Sets the users_to_remove of this SCIMMigrationResult.


        :param users_to_remove: The users_to_remove of this SCIMMigrationResult.  # noqa: E501
        :type: list[UserToRemove]
        """

        self._users_to_remove = users_to_remove

    @property
    def dry_run(self):
        """Gets the dry_run of this SCIMMigrationResult.  # noqa: E501


        :return: The dry_run of this SCIMMigrationResult.  # noqa: E501
        :rtype: bool
        """
        return self._dry_run

    @dry_run.setter
    def dry_run(self, dry_run):
        """Sets the dry_run of this SCIMMigrationResult.


        :param dry_run: The dry_run of this SCIMMigrationResult.  # noqa: E501
        :type: bool
        """

        self._dry_run = dry_run

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SCIMMigrationResult):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SCIMMigrationResult):
            return True

        return self.to_dict() != other.to_dict()



import contextlib
import os
import time
import webbrowser

import click

from anyscale.api import ApiClientWrapperInternal
from anyscale.cli_logger import BlockLogger
from anyscale.client import openapi_client
from anyscale.controllers.auth_controller import AuthController
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST


log = BlockLogger()


def get_unauthenticated_openapi_client():
    conf = openapi_client.Configuration(host=ANYSCALE_HOST)
    conf.proxy = os.environ.get("https_proxy")
    conf.connection_pool_maxsize = 100
    return openapi_client.DefaultApi(ApiClientWrapperInternal(conf))


@click.command(name="login", help="Log in to Anyscale using a URL.")
@click.option(
    "--no-expire", is_flag=True, default=False, help="Do not expire the token.",
)
@click.option(
    "--expire-in-days",
    type=int,
    default=7,
    help="Expire the token after this many days.",
)
@click.option(
    "--no-browser", is_flag=True, default=False, help="Do not open the browser.",
)
def anyscale_login(no_expire: bool, expire_in_days: int, no_browser: bool) -> None:
    """Log in to Anyscale using a URL
    This is the only unauthenticated API usage in the CLI."""
    if (  # noqa: SIM108 - inline comments explain magic values
        expire_in_days < 0 or no_expire
    ):
        duration_seconds = 0  # no expiration
    else:
        duration_seconds = expire_in_days * 24 * 60 * 60  # seconds

    api_client = get_unauthenticated_openapi_client()

    r = api_client.create_one_time_password_api_v2_users_create_otp_token_get(
        duration_seconds=duration_seconds
    ).result
    token = r.otp

    log.info("Please log in to Anyscale using the following URL:")
    log.info(r.url)

    # Open the URL in the browser. This will work on most platforms.
    # OK to suppress any uncaught exceptions, because the URL will be printed out anyway.
    if not no_browser:
        with contextlib.suppress(Exception):
            try:
                # Only attempt to open browser if it's not a CLI browser (which doesn't support JavaScript)
                if webbrowser.get().basename not in [
                    "www-browser",
                    "elinks",
                    "links",
                    "lynx",
                    "w3m",
                ]:
                    webbrowser.open_new_tab(r.url)
                else:
                    log.info(
                        "Could not open a JavaScript-capable web browser. Open the above URL in your browser manually."
                    )
            except webbrowser.Error:
                log.info(
                    "Could not open a web browser. Open the above URL in your browser manually."
                )

    # give user 3 minutes to log in (3 seconds per attempt)
    for _i in range(60):
        check_result = api_client.check_one_time_password_api_v2_users_request_otp_token_otp_get(
            token
        ).result
        if check_result.valid:
            AuthController().set(check_result.server_session_id)
            log.info("Successfully logged in to Anyscale.")
            if duration_seconds > 0:
                log.info(f"The authentication will expire in {expire_in_days} days.")
            return
        time.sleep(3)

    log.info("Failed to log in to Anyscale. Please try again.")


@click.command(
    name="logout", help="Log out from Anyscale. You CLI token will be revoked."
)
def anyscale_logout() -> None:
    AuthController().remove(ask=False)
    log.info("Successfully logged out from Anyscale.")



"""
Cloud File Storage Utilities

Common functions for verifying AWS EFS and GCP Filestore resources
across different cloud verification contexts.
"""

from typing import Dict, Optional

import boto3

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.cloud_deployment import CloudDeployment
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.file_storage import FileStorage
from anyscale.util import _get_aws_efs_mount_target_ip


class GCPFilestoreInfo:
    """Container for GCP Filestore information"""

    def __init__(
        self,
        filestore_id: str,
        exists: bool,
        mount_target_ip: Optional[str] = None,
        root_dir: Optional[str] = None,
        details: Optional[Dict] = None,
    ):
        self.filestore_id = filestore_id
        self.exists = exists
        self.mount_target_ip = mount_target_ip
        self.root_dir = root_dir
        self.details = details or {}


def _get_gcp_client_factory(logger, cloud_deployment: CloudDeployment):
    """Helper function to get GCP client factory."""
    from anyscale.cli_logger import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
        CloudSetupLogger,
    )
    from anyscale.utils.gcp_utils import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
        get_google_cloud_client_factory,
    )

    gcp_config = cloud_deployment.gcp_config or {}
    project_id = gcp_config.get("project_id")

    if not project_id:
        raise ValueError("No GCP project_id found in cloud deployment config")

    # Create GCP client factory - need to handle logger type mismatch
    if isinstance(logger, CloudSetupLogger):
        return get_google_cloud_client_factory(logger, project_id), project_id
    else:
        # Create a CloudSetupLogger wrapper for compatibility
        cloud_setup_logger = CloudSetupLogger()
        return (
            get_google_cloud_client_factory(cloud_setup_logger, project_id),
            project_id,
        )


def _get_detailed_filestore_config(
    factory, vpc_name: Optional[str], filestore_id: str, logger: BlockLogger
):
    """Helper function to get detailed filestore configuration."""
    if not vpc_name:
        return None, None

    try:
        from anyscale.cli_logger import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
            CloudSetupLogger,
        )
        from anyscale.utils import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
            gcp_utils,
        )

        cloud_setup_logger = CloudSetupLogger()
        filestore_config = gcp_utils.get_gcp_filestore_config_from_full_name(
            factory, vpc_name, filestore_id, cloud_setup_logger
        )
        if filestore_config:
            return filestore_config.mount_target_ip, filestore_config.root_dir
    except (ImportError, AttributeError, KeyError) as e:
        logger.warning(f"Could not retrieve detailed filestore config: {e}")

    return None, None


def get_gcp_filestore_info(
    filestore_id: str,
    cloud_deployment: CloudDeployment,
    logger: Optional[BlockLogger] = None,
) -> GCPFilestoreInfo:
    """
    Common function to fetch GCP Filestore information.

    Args:
        filestore_id: The GCP Filestore instance ID
        cloud_deployment: CloudDeployment object containing GCP config
        logger: Optional logger for output

    Returns:
        GCPFilestoreInfo object containing Filestore details and existence status
    """
    if logger is None:
        logger = BlockLogger()

    try:
        from anyscale.cli_logger import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
            CloudSetupLogger,
        )
        from anyscale.gcp_verification import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP deps loaded only when needed")
            GCPLogger,
            verify_filestore,
        )

        # Get GCP config and client factory
        factory, project_id = _get_gcp_client_factory(logger, cloud_deployment)

        gcp_config = cloud_deployment.gcp_config or {}
        vpc_name = gcp_config.get("vpc_name")

        if not vpc_name:
            logger.warning("No VPC name found - may affect connectivity checks")

        # Create GCP logger for filestore verification
        gcp_logger = GCPLogger(CloudSetupLogger(), project_id, None, True)

        # Use existing filestore verification logic
        region = cloud_deployment.region or ""
        verification_result = verify_filestore(
            factory=factory,
            file_store_instance_name=filestore_id,
            vpc_name=vpc_name,
            cloud_region=region,
            logger=gcp_logger,
            strict=False,
        )

        # Try to get detailed filestore config if available
        mount_target_ip, root_dir = _get_detailed_filestore_config(
            factory, vpc_name, filestore_id, logger
        )

        if verification_result:
            logger.info(f"GCP Filestore {filestore_id} exists and is accessible")
            return GCPFilestoreInfo(
                filestore_id=filestore_id,
                exists=True,
                mount_target_ip=mount_target_ip,
                root_dir=root_dir,
            )
        else:
            logger.error(f"GCP Filestore {filestore_id} verification failed")
            return GCPFilestoreInfo(filestore_id=filestore_id, exists=False)

    except ImportError as e:
        logger.warning(f"GCP verification dependencies not available: {e}")
        # Don't fail verification if GCP deps are missing
        return GCPFilestoreInfo(filestore_id=filestore_id, exists=True)
    except (ValueError, AttributeError, KeyError) as e:
        logger.error(f"Configuration error checking GCP Filestore {filestore_id}: {e}")
        return GCPFilestoreInfo(filestore_id=filestore_id, exists=False)


def verify_file_storage_exists(
    file_storage: FileStorage,
    cloud_deployment: CloudDeployment,
    logger: Optional[BlockLogger] = None,
) -> bool:
    """
    Generic function to verify file storage exists based on cloud provider.

    Args:
        file_storage: File storage configuration object
        cloud_deployment: CloudDeployment object
        logger: Optional logger for output

    Returns:
        bool: True if file storage exists, False otherwise
    """
    if logger is None:
        logger = BlockLogger()

    file_storage_id = getattr(file_storage, "file_storage_id", None)
    if not file_storage_id:
        logger.warning("No file_storage_id provided - skipping existence check")
        return True

    provider_value = (
        getattr(cloud_deployment.provider, "value", str(cloud_deployment.provider))
        if cloud_deployment.provider
        else None
    )

    if provider_value == CloudProviders.AWS:
        region = cloud_deployment.region
        if not region:
            logger.error("No region found for AWS EFS verification")
            return False

        boto3_session = boto3.Session(region_name=region)
        return _get_aws_efs_mount_target_ip(boto3_session, file_storage_id) is not None

    elif provider_value == CloudProviders.GCP:
        filestore_info = get_gcp_filestore_info(
            file_storage_id, cloud_deployment, logger
        )
        return filestore_info.exists

    else:
        logger.warning(
            f"File storage existence check not supported for provider: {provider_value}"
        )
        return True



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.client.openapi_client.models import (
    CloudDeploymentComputeConfig,
    ComputeNodeType,
    ComputeTemplateConfig,
    DecoratedComputeTemplate,
    DecoratedComputeTemplateConfig,
    PhysicalResources,
    Resources,
    WorkerNodeType,
)
from anyscale.cluster_compute import parse_cluster_compute_name_version
from anyscale.compute_config.models import (
    CloudDeployment,
    ComputeConfig,
    ComputeConfigListResult,
    ComputeConfigType,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    MultiResourceComputeConfig,
    WorkerNodeGroupConfig,
)
from anyscale.sdk.anyscale_client.models import ClusterComputeConfig


# Used to explicitly make the head node unschedulable.
# We can't leave resources empty because the backend will fill in CPU and GPU
# to match the instance type hardware.
UNSCHEDULABLE_RESOURCES = Resources(cpu=0, gpu=0)


class PrivateComputeConfigSDK(BaseSDK):
    def _convert_resource_dict_to_api_model(
        self, resource_dict: Optional[Dict[str, float]]
    ) -> Optional[Resources]:
        if resource_dict is None:
            return None

        resource_dict = deepcopy(resource_dict)
        return Resources(
            cpu=resource_dict.pop("CPU", None),
            gpu=resource_dict.pop("GPU", None),
            memory=resource_dict.pop("memory", None),
            object_store_memory=resource_dict.pop("object_store_memory", None),
            custom_resources=resource_dict or None,
        )

    def _convert_head_node_config_to_api_model(
        self,
        config: Union[None, Dict, HeadNodeConfig],
        *,
        cloud_id: str,
        cloud_resource_name: Optional[str] = None,
        schedulable_by_default: bool,
    ) -> ComputeNodeType:
        if config is None:
            cloud_resource_id = None
            if cloud_resource_name:
                cloud_resource = self._client.get_cloud_resource_by_name(
                    cloud_id=cloud_id, cloud_resource_name=cloud_resource_name
                )
                if cloud_resource:
                    cloud_resource_id = cloud_resource.cloud_resource_id

            # If no head node config is provided, use the cloud/cloud resource default.
            default: ClusterComputeConfig = self._client.get_default_compute_config(
                cloud_id=cloud_id, cloud_resource_id=cloud_resource_id,
            ).config

            api_model = ComputeNodeType(
                name="head-node",
                instance_type=default.head_node_type.instance_type,
                # Let the backend populate the physical resources
                # (regardless of what the default compute config says).
                resources=None if schedulable_by_default else UNSCHEDULABLE_RESOURCES,
                flags=default.flags,
            )
        else:
            # Make mypy happy.
            assert isinstance(config, HeadNodeConfig)

            flags: Dict[str, Any] = deepcopy(config.flags) if config.flags else {}
            if config.cloud_deployment:
                assert isinstance(config.cloud_deployment, CloudDeployment)
                flags["cloud_deployment"] = config.cloud_deployment.to_dict()

            # Convert required_resources to API model (OpenAPI PhysicalResources object)
            # Only pass non-None values to avoid sending null fields to backend
            required_resources_api = None
            if config.required_resources is not None:
                pr_dict = config.required_resources.to_dict()
                # Copy only the fields that are present in the dict
                # To add support for new fields, just add them to this list
                pr_kwargs = {
                    field: pr_dict[field]
                    for field in ["cpu", "gpu", "memory", "accelerator", "tpu",]
                    if field in pr_dict
                }
                # Handle tpu_hosts which maps to anyscale_tpu_hosts in the API
                if "tpu_hosts" in pr_dict:
                    pr_kwargs["anyscale_tpu_hosts"] = pr_dict["tpu_hosts"]
                required_resources_api = PhysicalResources(**pr_kwargs)

            api_model = ComputeNodeType(
                name="head-node",
                instance_type=config.instance_type,
                resources=self._convert_resource_dict_to_api_model(config.resources)
                if config.resources is not None or schedulable_by_default
                else UNSCHEDULABLE_RESOURCES,
                required_resources=required_resources_api,
                labels=config.labels,
                required_labels=config.required_labels,
                flags=flags or None,
                advanced_configurations_json=config.advanced_instance_config or None,
            )

        return api_model

    def _convert_worker_node_group_configs_to_api_models(
        self, configs: Optional[List[Union[Dict, WorkerNodeGroupConfig]]],
    ) -> Optional[List[WorkerNodeType]]:
        if configs is None:
            return None

        api_models = []
        for config in configs:
            # Make mypy happy.
            assert isinstance(config, WorkerNodeGroupConfig)

            flags: Dict[str, Any] = deepcopy(config.flags) if config.flags else {}
            if config.cloud_deployment:
                assert isinstance(config.cloud_deployment, CloudDeployment)
                flags["cloud_deployment"] = config.cloud_deployment.to_dict()

            # Convert required_resources to API model (OpenAPI PhysicalResources object)
            # Only pass non-None values to avoid sending null fields to backend
            required_resources_api = None
            if config.required_resources is not None:
                pr_dict = config.required_resources.to_dict()
                # Copy only the fields that are present in the dict
                # To add support for new fields, just add them to this list
                pr_kwargs = {
                    field: pr_dict[field]
                    for field in ["cpu", "gpu", "memory", "accelerator", "tpu",]
                    if field in pr_dict
                }
                # Handle tpu_hosts which maps to anyscale_tpu_hosts in the API
                if "tpu_hosts" in pr_dict:
                    pr_kwargs["anyscale_tpu_hosts"] = pr_dict["tpu_hosts"]
                required_resources_api = PhysicalResources(**pr_kwargs)

            api_model = WorkerNodeType(
                name=config.name,
                instance_type=config.instance_type,
                resources=self._convert_resource_dict_to_api_model(config.resources),
                required_resources=required_resources_api,
                labels=config.labels,
                required_labels=config.required_labels,
                min_workers=config.min_nodes,
                max_workers=config.max_nodes,
                use_spot=config.market_type
                in {MarketType.SPOT, MarketType.PREFER_SPOT},
                fallback_to_ondemand=config.market_type == MarketType.PREFER_SPOT,
                flags=flags or None,
                advanced_configurations_json=config.advanced_instance_config or None,
            )
            api_models.append(api_model)

        return api_models

    def _convert_single_deployment_compute_config_to_api_model(
        self,
        cloud_id: str,
        compute_config: ComputeConfig,
        base_flags: Optional[Dict[str, Any]] = None,
    ) -> CloudDeploymentComputeConfig:
        # We should only make the head node schedulable when it's the *only* node in the cluster.
        # `worker_nodes=None` uses the default serverless config, so this only happens if `worker_nodes`
        # is explicitly set to an empty list.

        # Merge the MultiResourceComputeConfig flags (which apply to all cloud resources) with the individual compute config flags.
        # The individual compute config flags will override the base flags.
        flags = deepcopy(base_flags) if base_flags else {}
        flags.update(compute_config.flags or {})
        flags["allow-cross-zone-autoscaling"] = compute_config.enable_cross_zone_scaling

        if compute_config.min_resources:
            flags["min_resources"] = compute_config.min_resources
        if compute_config.max_resources:
            flags["max_resources"] = compute_config.max_resources

        return CloudDeploymentComputeConfig(
            cloud_deployment=compute_config.cloud_resource,
            allowed_azs=compute_config.zones,
            head_node_type=self._convert_head_node_config_to_api_model(
                compute_config.head_node,
                cloud_id=cloud_id,
                cloud_resource_name=compute_config.cloud_resource,
                schedulable_by_default=(
                    not compute_config.worker_nodes
                    and not compute_config.auto_select_worker_config
                ),
            ),
            worker_node_types=self._convert_worker_node_group_configs_to_api_models(
                compute_config.worker_nodes,
            ),
            auto_select_worker_config=compute_config.auto_select_worker_config,
            flags=flags,
            advanced_configurations_json=compute_config.advanced_instance_config
            or None,
        )

    def create_compute_config(
        self, compute_config: ComputeConfigType, *, name: Optional[str] = None
    ) -> Tuple[str, str]:
        """Register the provided compute config and return its internal ID."""

        if name is not None:
            _, version = parse_cluster_compute_name_version(name)
            if version is not None:
                raise ValueError(
                    "A version tag cannot be provided when creating a compute config. "
                    "The latest version tag will be generated and returned."
                )

        if isinstance(compute_config, MultiResourceComputeConfig):
            return self.create_multi_deployment_compute_config(
                compute_config, name=name
            )
        else:
            assert isinstance(compute_config, ComputeConfig)
            return self.create_single_deployment_compute_config(
                compute_config, name=name
            )

    def create_single_deployment_compute_config(
        self, compute_config: ComputeConfig, *, name: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Register the provided single-deployment compute config and return its internal ID."""

        # Returns the default cloud if user-provided cloud is not specified (`None`).
        cloud_id = self.client.get_cloud_id(cloud_name=compute_config.cloud)  # type: ignore

        deployment_config = self._convert_single_deployment_compute_config_to_api_model(
            cloud_id, compute_config
        )

        compute_config_api_model = ComputeTemplateConfig(
            cloud_id=cloud_id,
            deployment_configs=[deployment_config],
            # For compatibility, continue setting the top-level fields.
            allowed_azs=deployment_config.allowed_azs,
            head_node_type=deployment_config.head_node_type,
            worker_node_types=deployment_config.worker_node_types,
            auto_select_worker_config=deployment_config.auto_select_worker_config,
            flags=deployment_config.flags,
            advanced_configurations_json=deployment_config.advanced_configurations_json
            or None,
        )

        full_name, compute_config_id = self.client.create_compute_config(
            compute_config_api_model, name=name
        )
        self.logger.info(f"Created compute config: '{full_name}'")
        ui_url = self.client.get_compute_config_ui_url(
            compute_config_id, cloud_id=cloud_id
        )
        self.logger.info(f"View the compute config in the UI: '{ui_url}'")
        return full_name, compute_config_id

    def create_multi_deployment_compute_config(
        self, compute_config: MultiResourceComputeConfig, *, name: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Register the provided multi-deployment compute config and return its internal ID."""
        # Returns the default cloud if user-provided cloud is not specified (`None`).
        cloud_id = self.client.get_cloud_id(cloud_name=compute_config.cloud)  # type: ignore

        # Convert each compute config to the CloudDeploymentComputeConfig API model.
        assert compute_config.configs
        deployment_configs = []
        for config in compute_config.configs:
            assert isinstance(config, ComputeConfig)
            deployment_configs.append(
                self._convert_single_deployment_compute_config_to_api_model(
                    cloud_id, config, compute_config.flags
                )
            )
        default_config = deployment_configs[0]

        compute_config_api_model = ComputeTemplateConfig(
            cloud_id=cloud_id,
            deployment_configs=deployment_configs,
            # For compatibility, use the first deployment config to set the top-level fields.
            allowed_azs=default_config.allowed_azs,
            head_node_type=default_config.head_node_type,
            worker_node_types=default_config.worker_node_types,
            auto_select_worker_config=default_config.auto_select_worker_config,
            flags=default_config.flags,
            advanced_configurations_json=default_config.advanced_configurations_json
            or None,
        )
        full_name, compute_config_id = self.client.create_compute_config(
            compute_config_api_model, name=name
        )
        self.logger.info(f"Created compute config: '{full_name}'")

        ui_url = self.client.get_compute_config_ui_url(
            compute_config_id, cloud_id=cloud_id
        )
        self.logger.info(f"View the compute config in the UI: '{ui_url}'")

        return full_name, compute_config_id

    def _convert_api_model_to_advanced_instance_config(
        self,
        api_model: Union[
            DecoratedComputeTemplateConfig, ComputeNodeType, WorkerNodeType
        ],
    ) -> Optional[Dict]:
        if api_model.advanced_configurations_json:
            return api_model.advanced_configurations_json

        # Only one of aws_advanced_configurations_json or gcp_advanced_configurations_json will be set.
        if api_model.aws_advanced_configurations_json:
            return api_model.aws_advanced_configurations_json
        if api_model.gcp_advanced_configurations_json:
            return api_model.gcp_advanced_configurations_json

        return None

    def _convert_api_model_to_resource_dict(
        self, resources: Optional[Resources]
    ) -> Optional[Dict[str, float]]:
        # Flatten the resource dict returned by the API and strip `None` values.
        if resources is None:
            return None

        return {
            k: v
            for k, v in {
                "CPU": resources.cpu,
                "GPU": resources.gpu,
                "memory": resources.memory,
                "object_store_memory": resources.object_store_memory,
                **(resources.custom_resources or {}),
            }.items()
            if v is not None
        }

    def _convert_api_model_to_head_node_config(
        self, api_model: ComputeNodeType
    ) -> HeadNodeConfig:
        flags: Dict[str, Any] = deepcopy(api_model.flags) or {}

        cloud_deployment_dict = flags.pop("cloud_deployment", None)
        cloud_deployment = (
            CloudDeployment.from_dict(cloud_deployment_dict)
            if cloud_deployment_dict
            else None
        )

        return HeadNodeConfig(
            instance_type=api_model.instance_type,
            resources=self._convert_api_model_to_resource_dict(api_model.resources),
            labels=api_model.labels,
            advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                api_model,
            ),
            flags=flags or None,
            cloud_deployment=cloud_deployment,
        )

    def _convert_api_models_to_worker_node_group_configs(
        self, api_models: List[WorkerNodeType]
    ) -> List[WorkerNodeGroupConfig]:
        # TODO(edoakes): support advanced_instance_config.
        configs = []
        for api_model in api_models:
            if api_model.use_spot and api_model.fallback_to_ondemand:
                market_type = MarketType.PREFER_SPOT
            elif api_model.use_spot:
                market_type = MarketType.SPOT
            else:
                market_type = MarketType.ON_DEMAND

            min_nodes = api_model.min_workers
            if min_nodes is None:
                min_nodes = 0

            max_nodes = api_model.max_workers
            if max_nodes is None:
                # TODO(edoakes): this defaulting to 10 seems like really strange
                # behavior here but I'm copying what the UI does. In Shomil's new
                # API let's not make these optional.
                max_nodes = 10

            flags: Dict[str, Any] = deepcopy(api_model.flags) or {}

            cloud_deployment_dict = flags.pop("cloud_deployment", None)
            cloud_deployment = (
                CloudDeployment.from_dict(cloud_deployment_dict)
                if cloud_deployment_dict
                else None
            )

            configs.append(
                WorkerNodeGroupConfig(
                    name=api_model.name,
                    instance_type=api_model.instance_type,
                    resources=self._convert_api_model_to_resource_dict(
                        api_model.resources
                    ),
                    labels=api_model.labels,
                    advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                        api_model,
                    ),
                    min_nodes=min_nodes,
                    max_nodes=max_nodes,
                    market_type=market_type,
                    flags=flags or None,
                    cloud_deployment=cloud_deployment,
                )
            )

        return configs

    def _convert_cloud_deployment_compute_config_api_model_to_single_resource_compute_config(
        self, cloud_name: str, api_model: CloudDeploymentComputeConfig,
    ) -> ComputeConfig:
        worker_nodes = None
        if not api_model.auto_select_worker_config:
            if api_model.worker_node_types is not None:
                # Convert worker node types when they are present.
                worker_nodes = self._convert_api_models_to_worker_node_group_configs(
                    api_model.worker_node_types
                )
            else:
                # An explicit head-node-only cluster (no worker nodes configured).
                worker_nodes = []

        zones = None
        # NOTE(edoakes): the API returns '["any"]' if no AZs are passed in on the creation path.
        if api_model.allowed_azs not in [["any"], []]:
            zones = api_model.allowed_azs

        enable_cross_zone_scaling = False
        flags: Dict[str, Any] = deepcopy(api_model.flags) or {}
        enable_cross_zone_scaling = flags.pop("allow-cross-zone-autoscaling", False)
        min_resources = flags.pop("min_resources", None)
        max_resources = flags.pop("max_resources", None)
        if max_resources is None:
            max_resources = {}
            max_cpus = flags.pop("max-cpus", None)
            if max_cpus:
                max_resources["CPU"] = max_cpus
            max_gpus = flags.pop("max-gpus", None)
            if max_gpus:
                max_resources["GPU"] = max_gpus

        return ComputeConfig(
            cloud=cloud_name,
            cloud_resource=api_model.cloud_deployment,
            zones=zones,
            advanced_instance_config=api_model.advanced_configurations_json or None,
            enable_cross_zone_scaling=enable_cross_zone_scaling,
            head_node=self._convert_api_model_to_head_node_config(
                api_model.head_node_type
            ),
            worker_nodes=worker_nodes,
            min_resources=min_resources,
            max_resources=max_resources or None,
            flags=flags,
            auto_select_worker_config=api_model.auto_select_worker_config or False,
        )

    def _convert_api_model_to_compute_config_version(
        self, api_model: DecoratedComputeTemplate  # noqa: ARG002
    ) -> ComputeConfigVersion:
        api_model_config: DecoratedComputeTemplateConfig = api_model.config
        cloud = self.client.get_cloud(cloud_id=api_model_config.cloud_id)
        if cloud is None:
            raise RuntimeError(
                f"Cloud with ID '{api_model_config.cloud_id}' not found. "
                "This should never happen; please reach out to Anyscale support."
            )

        configs = None
        if api_model_config.deployment_configs:
            configs = [
                self._convert_cloud_deployment_compute_config_api_model_to_single_resource_compute_config(
                    cloud.name, config
                )
                for config in api_model_config.deployment_configs
            ]
            if len(configs) == 1:
                # If there's only one deployment config, return it directly.
                return ComputeConfigVersion(
                    name=f"{api_model.name}:{api_model.version}",
                    id=api_model.id,
                    config=configs[0],
                )
            return ComputeConfigVersion(
                name=f"{api_model.name}:{api_model.version}",
                id=api_model.id,
                config=MultiResourceComputeConfig(cloud=cloud.name, configs=configs),
            )

        # If there are no deployment configs, this is a compute config for a single cloud deployment - parse the top-level fields.

        worker_nodes = None
        if not api_model_config.auto_select_worker_config:
            if api_model_config.worker_node_types is not None:
                # Convert worker node types when they are present.
                worker_nodes = self._convert_api_models_to_worker_node_group_configs(
                    api_model_config.worker_node_types
                )
            else:
                # An explicit head-node-only cluster (no worker nodes configured).
                worker_nodes = []

        zones = None
        # NOTE(edoakes): the API returns '["any"]' if no AZs are passed in on the creation path.
        if api_model_config.allowed_azs not in [["any"], []]:
            zones = api_model_config.allowed_azs

        enable_cross_zone_scaling = False
        flags: Dict[str, Any] = deepcopy(api_model_config.flags) or {}
        enable_cross_zone_scaling = flags.pop("allow-cross-zone-autoscaling", False)
        min_resources = flags.pop("min_resources", None)
        max_resources = flags.pop("max_resources", None)
        if max_resources is None:
            max_resources = {}
            max_cpus = flags.pop("max-cpus", None)
            if max_cpus:
                max_resources["CPU"] = max_cpus
            max_gpus = flags.pop("max-gpus", None)
            if max_gpus:
                max_resources["GPU"] = max_gpus

        return ComputeConfigVersion(
            name=f"{api_model.name}:{api_model.version}",
            id=api_model.id,
            config=ComputeConfig(
                cloud=cloud.name,
                zones=zones,
                advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                    api_model_config
                ),
                enable_cross_zone_scaling=enable_cross_zone_scaling,
                head_node=self._convert_api_model_to_head_node_config(
                    api_model_config.head_node_type
                ),
                worker_nodes=worker_nodes,  # type: ignore
                min_resources=min_resources,
                max_resources=max_resources or None,
                auto_select_worker_config=api_model_config.auto_select_worker_config,
                flags=flags,
            ),
        )

    def _resolve_id(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        include_archived: bool = False,
    ) -> str:
        if id is not None:
            compute_config_id = id
        elif name is not None:
            compute_config_id = self.client.get_compute_config_id(
                compute_config_name=name, cloud=cloud, include_archived=include_archived
            )
            if compute_config_id is None:
                raise RuntimeError(f"Compute config '{name}' not found.")
        else:
            raise ValueError("Either name or ID must be provided.")

        return compute_config_id

    def get_compute_config(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        include_archived: bool = False,
    ) -> ComputeConfigVersion:
        """Get the compute config for the provided name.

        The name can contain an optional version, e.g., '<name>:<version>'.
        If no version is provided, the latest one will be returned.
        """
        compute_config_id = self._resolve_id(
            id=id, name=name, cloud=cloud, include_archived=include_archived
        )
        compute_config = self.client.get_compute_config(compute_config_id)
        if compute_config is None:
            raise RuntimeError(
                f"Compute config with ID '{compute_config_id}' not found.'"
            )
        return self._convert_api_model_to_compute_config_version(compute_config)

    def archive_compute_config(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
    ):
        compute_config_id = self._resolve_id(id=id, name=name, cloud=cloud)
        self.client.archive_compute_config(compute_config_id=compute_config_id)
        self.logger.info("Compute config is successfully archived.")

    def get_default_compute_config(
        self, *, cloud: Optional[str] = None, cloud_resource: Optional[str] = None,
    ) -> ComputeConfigVersion:
        """Get the default compute config for the specified cloud.

        Args:
            cloud: Name of the cloud. If not provided, uses the default cloud.
            cloud_resource: Name of the cloud resource. If not provided, uses the
                default cloud resource for the cloud.

        Returns:
            ComputeConfigVersion containing the default compute config.
        """
        # Resolve cloud name to cloud ID
        cloud_id = self.client.get_cloud_id(cloud_name=cloud)  # type: ignore

        # Resolve cloud resource name to cloud resource ID if provided
        cloud_resource_id = None
        if cloud_resource:
            cloud_resource_obj = self.client.get_cloud_resource_by_name(
                cloud_id=cloud_id, cloud_resource_name=cloud_resource
            )
            if cloud_resource_obj:
                cloud_resource_id = cloud_resource_obj.cloud_resource_id

        # Get the default compute config from the client
        default_config = self.client.get_default_compute_config(
            cloud_id=cloud_id, cloud_resource_id=cloud_resource_id
        )

        # Convert to SDK model
        return self._convert_api_model_to_compute_config_version(default_config)

    def list_compute_configs(  # noqa: PLR0913
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud_id: Optional[str] = None,
        cloud_name: Optional[str] = None,
        sort_by: str = "last_modified_at",
        sort_order: str = "asc",
        max_items: int = 20,
        next_token: Optional[str] = None,
        include_shared: bool = False,
    ):
        """List compute configs with filtering, sorting, and pagination.

        Args:
            name: Filter by compute config name
            id: Filter by specific compute config ID
            cloud_id: Filter by cloud ID
            cloud_name: Filter by cloud name (will be resolved to cloud_id)
            sort_by: Field to sort by (name, created_at, last_modified_at)
            sort_order: Sort order (asc or desc)
            max_items: Maximum number of items to return per page
            next_token: Pagination token for fetching next page
            include_shared: Include configs shared with the user

        Returns:
            ComputeConfigListResult with:
            - results: List of compute config objects
            - next_token: Pagination token for next page (None if no more results)
            - count: Number of results returned
        """

        # If fetching by ID, return single result
        if id:
            compute_config = self.client.get_compute_config(id)
            if compute_config is None:
                raise RuntimeError(f"Compute config with ID '{id}' not found.")
            return ComputeConfigListResult(
                results=[compute_config], next_token=None, count=1
            )

        # Resolve cloud_name to cloud_id if provided
        resolved_cloud_id = cloud_id
        if cloud_name:
            # Directly use the client's get_cloud_id method
            resolved_cloud_id = self.client.get_cloud_id(cloud_name=cloud_name)  # type: ignore

        # Build query for search endpoint
        query: Dict[str, Any] = {"paging": {"count": max_items}}

        # Add name filter if specified
        if name:
            query["name"] = {"equals": name}

        # Add creator_id filter if not including shared configs
        if not include_shared and not name:
            # Use self.client instead of creating new BaseController
            user_info = self.client.get_user_info()  # type: ignore
            query["creator_id"] = user_info["id"] if isinstance(user_info, dict) else user_info.id  # type: ignore

        # Add cloud filter if specified
        if resolved_cloud_id:
            query["cloud_id"] = resolved_cloud_id

        # Add pagination token if provided
        if next_token:
            query["paging"]["paging_token"] = next_token

        # Add server-side sorting
        sort_field_map = {
            "name": "NAME",
            "created_at": "CREATED_AT",
            "last_modified_at": "LAST_MODIFIED_AT",
        }
        query["sort_by_clauses"] = [
            {
                "sort_field": sort_field_map.get(sort_by, "LAST_MODIFIED_AT"),
                "sort_order": sort_order.upper(),
            }
        ]

        # Make API call
        response = self.client.search_cluster_computes(query)  # type: ignore

        # Return structured result
        pagination_token = (
            response.metadata.next_paging_token if response.metadata else None
        )
        return ComputeConfigListResult(
            results=response.results,
            next_token=pagination_token,
            count=len(response.results),
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class TableRedactionResult(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'table_name': 'str',
        'total_rows': 'int',
        'redacted_rows': 'int',
        'status': 'str',
        'error_message': 'str'
    }

    attribute_map = {
        'table_name': 'table_name',
        'total_rows': 'total_rows',
        'redacted_rows': 'redacted_rows',
        'status': 'status',
        'error_message': 'error_message'
    }

    def __init__(self, table_name=None, total_rows=None, redacted_rows=None, status=None, error_message=None, local_vars_configuration=None):  # noqa: E501
        """TableRedactionResult - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._table_name = None
        self._total_rows = None
        self._redacted_rows = None
        self._status = None
        self._error_message = None
        self.discriminator = None

        self.table_name = table_name
        self.total_rows = total_rows
        self.redacted_rows = redacted_rows
        self.status = status
        if error_message is not None:
            self.error_message = error_message

    @property
    def table_name(self):
        """Gets the table_name of this TableRedactionResult.  # noqa: E501


        :return: The table_name of this TableRedactionResult.  # noqa: E501
        :rtype: str
        """
        return self._table_name

    @table_name.setter
    def table_name(self, table_name):
        """Sets the table_name of this TableRedactionResult.


        :param table_name: The table_name of this TableRedactionResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and table_name is None:  # noqa: E501
            raise ValueError("Invalid value for `table_name`, must not be `None`")  # noqa: E501

        self._table_name = table_name

    @property
    def total_rows(self):
        """Gets the total_rows of this TableRedactionResult.  # noqa: E501


        :return: The total_rows of this TableRedactionResult.  # noqa: E501
        :rtype: int
        """
        return self._total_rows

    @total_rows.setter
    def total_rows(self, total_rows):
        """Sets the total_rows of this TableRedactionResult.


        :param total_rows: The total_rows of this TableRedactionResult.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and total_rows is None:  # noqa: E501
            raise ValueError("Invalid value for `total_rows`, must not be `None`")  # noqa: E501

        self._total_rows = total_rows

    @property
    def redacted_rows(self):
        """Gets the redacted_rows of this TableRedactionResult.  # noqa: E501


        :return: The redacted_rows of this TableRedactionResult.  # noqa: E501
        :rtype: int
        """
        return self._redacted_rows

    @redacted_rows.setter
    def redacted_rows(self, redacted_rows):
        """Sets the redacted_rows of this TableRedactionResult.


        :param redacted_rows: The redacted_rows of this TableRedactionResult.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and redacted_rows is None:  # noqa: E501
            raise ValueError("Invalid value for `redacted_rows`, must not be `None`")  # noqa: E501

        self._redacted_rows = redacted_rows

    @property
    def status(self):
        """Gets the status of this TableRedactionResult.  # noqa: E501


        :return: The status of this TableRedactionResult.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this TableRedactionResult.


        :param status: The status of this TableRedactionResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def error_message(self):
        """Gets the error_message of this TableRedactionResult.  # noqa: E501


        :return: The error_message of this TableRedactionResult.  # noqa: E501
        :rtype: str
        """
        return self._error_message

    @error_message.setter
    def error_message(self, error_message):
        """Sets the error_message of this TableRedactionResult.


        :param error_message: The error_message of this TableRedactionResult.  # noqa: E501
        :type: str
        """

        self._error_message = error_message

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TableRedactionResult):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, TableRedactionResult):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class NodeMetrics(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'node_id': 'str',
        'cpu_usage': 'float',
        'memory_used': 'int',
        'memory_total': 'int',
        'gpus': 'list[GPUUsage]',
        'object_store_usage': 'int',
        'object_store_free': 'int',
        'disk_usage': 'int',
        'disk_free': 'int',
        'network_sent': 'int',
        'network_received': 'int'
    }

    attribute_map = {
        'node_id': 'node_id',
        'cpu_usage': 'cpu_usage',
        'memory_used': 'memory_used',
        'memory_total': 'memory_total',
        'gpus': 'gpus',
        'object_store_usage': 'object_store_usage',
        'object_store_free': 'object_store_free',
        'disk_usage': 'disk_usage',
        'disk_free': 'disk_free',
        'network_sent': 'network_sent',
        'network_received': 'network_received'
    }

    def __init__(self, node_id=None, cpu_usage=None, memory_used=None, memory_total=None, gpus=None, object_store_usage=None, object_store_free=None, disk_usage=None, disk_free=None, network_sent=None, network_received=None, local_vars_configuration=None):  # noqa: E501
        """NodeMetrics - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._node_id = None
        self._cpu_usage = None
        self._memory_used = None
        self._memory_total = None
        self._gpus = None
        self._object_store_usage = None
        self._object_store_free = None
        self._disk_usage = None
        self._disk_free = None
        self._network_sent = None
        self._network_received = None
        self.discriminator = None

        self.node_id = node_id
        if cpu_usage is not None:
            self.cpu_usage = cpu_usage
        if memory_used is not None:
            self.memory_used = memory_used
        if memory_total is not None:
            self.memory_total = memory_total
        self.gpus = gpus
        if object_store_usage is not None:
            self.object_store_usage = object_store_usage
        if object_store_free is not None:
            self.object_store_free = object_store_free
        if disk_usage is not None:
            self.disk_usage = disk_usage
        if disk_free is not None:
            self.disk_free = disk_free
        if network_sent is not None:
            self.network_sent = network_sent
        if network_received is not None:
            self.network_received = network_received

    @property
    def node_id(self):
        """Gets the node_id of this NodeMetrics.  # noqa: E501

        The ID of the node.  # noqa: E501

        :return: The node_id of this NodeMetrics.  # noqa: E501
        :rtype: str
        """
        return self._node_id

    @node_id.setter
    def node_id(self, node_id):
        """Sets the node_id of this NodeMetrics.

        The ID of the node.  # noqa: E501

        :param node_id: The node_id of this NodeMetrics.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and node_id is None:  # noqa: E501
            raise ValueError("Invalid value for `node_id`, must not be `None`")  # noqa: E501

        self._node_id = node_id

    @property
    def cpu_usage(self):
        """Gets the cpu_usage of this NodeMetrics.  # noqa: E501

        The CPU usage in percentage of the node.  # noqa: E501

        :return: The cpu_usage of this NodeMetrics.  # noqa: E501
        :rtype: float
        """
        return self._cpu_usage

    @cpu_usage.setter
    def cpu_usage(self, cpu_usage):
        """Sets the cpu_usage of this NodeMetrics.

        The CPU usage in percentage of the node.  # noqa: E501

        :param cpu_usage: The cpu_usage of this NodeMetrics.  # noqa: E501
        :type: float
        """

        self._cpu_usage = cpu_usage

    @property
    def memory_used(self):
        """Gets the memory_used of this NodeMetrics.  # noqa: E501

        The memory used in bytes of the node.  # noqa: E501

        :return: The memory_used of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._memory_used

    @memory_used.setter
    def memory_used(self, memory_used):
        """Sets the memory_used of this NodeMetrics.

        The memory used in bytes of the node.  # noqa: E501

        :param memory_used: The memory_used of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._memory_used = memory_used

    @property
    def memory_total(self):
        """Gets the memory_total of this NodeMetrics.  # noqa: E501

        The memory total in bytes of the node.  # noqa: E501

        :return: The memory_total of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._memory_total

    @memory_total.setter
    def memory_total(self, memory_total):
        """Sets the memory_total of this NodeMetrics.

        The memory total in bytes of the node.  # noqa: E501

        :param memory_total: The memory_total of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._memory_total = memory_total

    @property
    def gpus(self):
        """Gets the gpus of this NodeMetrics.  # noqa: E501

        The GPUs used by the node.  # noqa: E501

        :return: The gpus of this NodeMetrics.  # noqa: E501
        :rtype: list[GPUUsage]
        """
        return self._gpus

    @gpus.setter
    def gpus(self, gpus):
        """Sets the gpus of this NodeMetrics.

        The GPUs used by the node.  # noqa: E501

        :param gpus: The gpus of this NodeMetrics.  # noqa: E501
        :type: list[GPUUsage]
        """
        if self.local_vars_configuration.client_side_validation and gpus is None:  # noqa: E501
            raise ValueError("Invalid value for `gpus`, must not be `None`")  # noqa: E501

        self._gpus = gpus

    @property
    def object_store_usage(self):
        """Gets the object_store_usage of this NodeMetrics.  # noqa: E501

        The object store usage in bytes of the node.  # noqa: E501

        :return: The object_store_usage of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._object_store_usage

    @object_store_usage.setter
    def object_store_usage(self, object_store_usage):
        """Sets the object_store_usage of this NodeMetrics.

        The object store usage in bytes of the node.  # noqa: E501

        :param object_store_usage: The object_store_usage of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._object_store_usage = object_store_usage

    @property
    def object_store_free(self):
        """Gets the object_store_free of this NodeMetrics.  # noqa: E501

        The object store free in bytes of the node.  # noqa: E501

        :return: The object_store_free of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._object_store_free

    @object_store_free.setter
    def object_store_free(self, object_store_free):
        """Sets the object_store_free of this NodeMetrics.

        The object store free in bytes of the node.  # noqa: E501

        :param object_store_free: The object_store_free of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._object_store_free = object_store_free

    @property
    def disk_usage(self):
        """Gets the disk_usage of this NodeMetrics.  # noqa: E501

        The disk usage in bytes of the node.  # noqa: E501

        :return: The disk_usage of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._disk_usage

    @disk_usage.setter
    def disk_usage(self, disk_usage):
        """Sets the disk_usage of this NodeMetrics.

        The disk usage in bytes of the node.  # noqa: E501

        :param disk_usage: The disk_usage of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._disk_usage = disk_usage

    @property
    def disk_free(self):
        """Gets the disk_free of this NodeMetrics.  # noqa: E501

        The disk free in bytes of the node.  # noqa: E501

        :return: The disk_free of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._disk_free

    @disk_free.setter
    def disk_free(self, disk_free):
        """Sets the disk_free of this NodeMetrics.

        The disk free in bytes of the node.  # noqa: E501

        :param disk_free: The disk_free of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._disk_free = disk_free

    @property
    def network_sent(self):
        """Gets the network_sent of this NodeMetrics.  # noqa: E501

        The network sent in bytes of the node.  # noqa: E501

        :return: The network_sent of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._network_sent

    @network_sent.setter
    def network_sent(self, network_sent):
        """Sets the network_sent of this NodeMetrics.

        The network sent in bytes of the node.  # noqa: E501

        :param network_sent: The network_sent of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._network_sent = network_sent

    @property
    def network_received(self):
        """Gets the network_received of this NodeMetrics.  # noqa: E501

        The network received in bytes of the node.  # noqa: E501

        :return: The network_received of this NodeMetrics.  # noqa: E501
        :rtype: int
        """
        return self._network_received

    @network_received.setter
    def network_received(self, network_received):
        """Sets the network_received of this NodeMetrics.

        The network received in bytes of the node.  # noqa: E501

        :param network_received: The network_received of this NodeMetrics.  # noqa: E501
        :type: int
        """

        self._network_received = network_received

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, NodeMetrics):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, NodeMetrics):
            return True

        return self.to_dict() != other.to_dict()
